from botocore.exceptions import ClientError
import boto3
import sure  # noqa # pylint: disable=unused-import
import pytest
from moto import mock_elastictranscoder
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID


@mock_elastictranscoder
def test_create_simple_pipeline():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    role = create_role_name("nonexistingrole")
    response = client.create_pipeline(
        Name="testpipeline",
        InputBucket="inputtest",
        OutputBucket="outputtest",
        Role=role,
    )
    response["ResponseMetadata"]["HTTPStatusCode"].should.equal(201)

    pipeline = response["Pipeline"]
    pipeline.should.have.key("Id")
    pipeline.should.have.key("Name").being.equal("testpipeline")
    pipeline.should.have.key("Arn").being.equal(
        f"arn:aws:elastictranscoder:{region}:{ACCOUNT_ID}:pipeline/{pipeline['Id']}"
    )
    pipeline.should.have.key("Status").being.equal("Active")
    pipeline.should.have.key("InputBucket").being.equal("inputtest")
    pipeline.should.have.key("OutputBucket").being.equal("outputtest")
    pipeline.should.have.key("Role").being.equal(role)
    pipeline.should.have.key("Notifications").being.equal(
        {"Progressing": "", "Completed": "", "Warning": "", "Error": ""}
    )
    pipeline.should.have.key("ContentConfig")
    pipeline["ContentConfig"].should.have.key("Bucket").being.equal("outputtest")
    pipeline["ContentConfig"].should.have.key("Permissions").being.equal([])
    pipeline.should.have.key("ThumbnailConfig")
    pipeline["ThumbnailConfig"].should.have.key("Bucket").being.equal("outputtest")
    pipeline["ThumbnailConfig"].should.have.key("Permissions").being.equal([])

    response.should.have.key("Warnings").being.equal([])


@mock_elastictranscoder
def test_create_pipeline_with_content_config():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    role = create_role_name("nonexistingrole")
    response = client.create_pipeline(
        Name="testpipeline",
        InputBucket="inputtest",
        ContentConfig={"Bucket": "outputtest"},
        ThumbnailConfig={"Bucket": "outputtest"},
        Role=role,
    )
    response["ResponseMetadata"]["HTTPStatusCode"].should.equal(201)

    pipeline = response["Pipeline"]
    pipeline.should.have.key("Id")
    pipeline.should.have.key("Name").being.equal("testpipeline")
    pipeline.should.have.key("Arn").being.equal(
        f"arn:aws:elastictranscoder:{region}:{ACCOUNT_ID}:pipeline/{pipeline['Id']}"
    )
    pipeline.should.have.key("Status").being.equal("Active")
    pipeline.should.have.key("InputBucket").being.equal("inputtest")
    pipeline.should.have.key("OutputBucket").being.equal("outputtest")
    pipeline.should.have.key("Role").being.equal(role)
    pipeline.should.have.key("Notifications").being.equal(
        {"Progressing": "", "Completed": "", "Warning": "", "Error": ""}
    )
    pipeline.should.have.key("ContentConfig")
    pipeline["ContentConfig"].should.have.key("Bucket").being.equal("outputtest")
    pipeline["ContentConfig"].should.have.key("Permissions").being.equal([])
    pipeline.should.have.key("ThumbnailConfig")
    pipeline["ThumbnailConfig"].should.have.key("Bucket").being.equal("outputtest")
    pipeline["ThumbnailConfig"].should.have.key("Permissions").being.equal([])


@mock_elastictranscoder
def test_create_pipeline_with_outputbucket_and_content_config():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    role = create_role_name("nonexistingrole")
    with pytest.raises(ClientError) as ex:
        client.create_pipeline(
            Name="testpipeline",
            InputBucket="inputtest",
            OutputBucket="outputtest",
            ContentConfig={"Bucket": "configoutputtest"},
            Role=role,
        )
    err = ex.value.response["Error"]
    err["Code"].should.equal("ValidationException")
    err["Message"].should.equal(
        "[OutputBucket and ContentConfig are mutually exclusive.]"
    )


@mock_elastictranscoder
def test_create_pipeline_without_thumbnail_config():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    role = create_role_name("nonexistingrole")
    with pytest.raises(ClientError) as ex:
        client.create_pipeline(
            Name="testpipeline",
            InputBucket="inputtest",
            ContentConfig={"Bucket": "outputtest"},
            Role=role,
        )
    err = ex.value.response["Error"]
    err["Code"].should.equal("ValidationException")
    err["Message"].should.equal(
        "[ThumbnailConfig:Bucket is not allowed to be null if ContentConfig is specified.]"
    )


@mock_elastictranscoder
def test_create_pipeline_without_role():
    client = boto3.client("elastictranscoder", region_name="us-east-1")

    with pytest.raises(ClientError) as ex:
        client.create_pipeline(Name="testpipeline", InputBucket="inputtest", Role="")
    err = ex.value.response["Error"]
    err["Code"].should.equal("ValidationException")
    err["Message"].should.equal("Role cannot be blank")


@mock_elastictranscoder
def test_create_pipeline_with_invalid_role():
    client = boto3.client("elastictranscoder", region_name="us-east-1")

    with pytest.raises(ClientError) as ex:
        client.create_pipeline(
            Name="testpipeline", InputBucket="inputtest", Role="asdf"
        )
    err = ex.value.response["Error"]
    err["Code"].should.equal("ValidationException")
    err["Message"].should.equal("Role ARN is invalid: asdf")


@mock_elastictranscoder
def test_create_pipeline_without_output():
    client = boto3.client("elastictranscoder", region_name="us-east-1")

    with pytest.raises(ClientError) as ex:
        client.create_pipeline(
            Name="testpipeline",
            InputBucket="inputtest",
            Role=create_role_name("nonexistingrole"),
        )
    err = ex.value.response["Error"]
    err["Code"].should.equal("ValidationException")
    err["Message"].should.equal(
        "[OutputBucket and ContentConfig:Bucket are not allowed to both be null.]"
    )


@mock_elastictranscoder
def test_list_pipelines():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    role = create_role_name("nonexistingrole")
    client.create_pipeline(
        Name="testpipeline",
        InputBucket="inputtest",
        OutputBucket="outputtest",
        Role=role,
    )

    response = client.list_pipelines()
    response["ResponseMetadata"]["HTTPStatusCode"].should.equal(200)
    response.should.have.key("Pipelines").being.length_of(1)

    pipeline = response["Pipelines"][0]
    pipeline.should.have.key("Id")
    pipeline.should.have.key("Name").being.equal("testpipeline")
    pipeline.should.have.key("Arn").being.equal(
        f"arn:aws:elastictranscoder:{region}:{ACCOUNT_ID}:pipeline/{pipeline['Id']}"
    )
    pipeline.should.have.key("Status").being.equal("Active")
    pipeline.should.have.key("InputBucket").being.equal("inputtest")
    pipeline.should.have.key("OutputBucket").being.equal("outputtest")
    pipeline.should.have.key("Role").being.equal(role)
    pipeline.should.have.key("Notifications").being.equal(
        {"Progressing": "", "Completed": "", "Warning": "", "Error": ""}
    )
    pipeline.should.have.key("ContentConfig")
    pipeline["ContentConfig"].should.have.key("Bucket").being.equal("outputtest")
    pipeline["ContentConfig"].should.have.key("Permissions").being.equal([])
    pipeline.should.have.key("ThumbnailConfig")
    pipeline["ThumbnailConfig"].should.have.key("Bucket").being.equal("outputtest")
    pipeline["ThumbnailConfig"].should.have.key("Permissions").being.equal([])


@mock_elastictranscoder
def test_read_pipeline():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    role = create_role_name("nonexistingrole")
    pipeline = client.create_pipeline(
        Name="testpipeline",
        InputBucket="inputtest",
        OutputBucket="outputtest",
        Role=role,
    )["Pipeline"]

    response = client.read_pipeline(Id=pipeline["Id"])

    response["ResponseMetadata"]["HTTPStatusCode"].should.equal(200)
    response.should.have.key("Pipeline")

    pipeline = response["Pipeline"]
    pipeline.should.have.key("Id")
    pipeline.should.have.key("Name").being.equal("testpipeline")
    pipeline.should.have.key("Arn").being.equal(
        f"arn:aws:elastictranscoder:{region}:{ACCOUNT_ID}:pipeline/{pipeline['Id']}"
    )
    pipeline.should.have.key("Status").being.equal("Active")
    pipeline.should.have.key("InputBucket").being.equal("inputtest")
    pipeline.should.have.key("OutputBucket").being.equal("outputtest")
    pipeline.should.have.key("Role").being.equal(role)
    pipeline.should.have.key("Notifications").being.equal(
        {"Progressing": "", "Completed": "", "Warning": "", "Error": ""}
    )
    pipeline.should.have.key("ContentConfig")
    pipeline["ContentConfig"].should.have.key("Bucket").being.equal("outputtest")
    pipeline["ContentConfig"].should.have.key("Permissions").being.equal([])
    pipeline.should.have.key("ThumbnailConfig")
    pipeline["ThumbnailConfig"].should.have.key("Bucket").being.equal("outputtest")
    pipeline["ThumbnailConfig"].should.have.key("Permissions").being.equal([])


@mock_elastictranscoder
def test_read_unknown_pipeline_format():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    with pytest.raises(ClientError) as ex:
        client.read_pipeline(Id="unknown-pipeline")
    err = ex.value.response["Error"]
    err["Code"].should.equal("ValidationException")
    err["Message"].should.equal(
        "1 validation error detected: Value 'unknown-pipeline' at 'id' failed to satisfy constraint: Member must satisfy regular expression pattern: ^\\d{13}-\\w{6}$"
    )


@mock_elastictranscoder
def test_read_nonexisting_pipeline_format():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    pipeline_id = "0000000000000-abcdef"
    with pytest.raises(ClientError) as ex:
        client.read_pipeline(Id=pipeline_id)
    err = ex.value.response["Error"]
    err["Code"].should.equal("ResourceNotFoundException")
    err["Message"].should.equal(
        f"The specified pipeline was not found: account={ACCOUNT_ID}, pipelineId={pipeline_id}."
    )


@mock_elastictranscoder
def test_update_pipeline_name():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    role = create_role_name("nonexistingrole")
    pipeline = client.create_pipeline(
        Name="testpipeline",
        InputBucket="inputtest",
        OutputBucket="outputtest",
        Role=role,
    )["Pipeline"]
    response = client.update_pipeline(Id=pipeline["Id"], Name="newtestpipeline")

    response["ResponseMetadata"]["HTTPStatusCode"].should.equal(200)
    response.should.have.key("Pipeline")

    pipeline = response["Pipeline"]
    pipeline.should.have.key("Id")
    pipeline.should.have.key("Name").being.equal("newtestpipeline")
    pipeline.should.have.key("Arn").being.equal(
        f"arn:aws:elastictranscoder:{region}:{ACCOUNT_ID}:pipeline/{pipeline['Id']}"
    )
    pipeline.should.have.key("Status").being.equal("Active")
    pipeline.should.have.key("InputBucket").being.equal("inputtest")
    pipeline.should.have.key("OutputBucket").being.equal("outputtest")
    pipeline.should.have.key("Role").being.equal(role)
    pipeline.should.have.key("Notifications").being.equal(
        {"Progressing": "", "Completed": "", "Warning": "", "Error": ""}
    )
    pipeline.should.have.key("ContentConfig")
    pipeline["ContentConfig"].should.have.key("Bucket").being.equal("outputtest")
    pipeline["ContentConfig"].should.have.key("Permissions").being.equal([])
    pipeline.should.have.key("ThumbnailConfig")
    pipeline["ThumbnailConfig"].should.have.key("Bucket").being.equal("outputtest")
    pipeline["ThumbnailConfig"].should.have.key("Permissions").being.equal([])


@mock_elastictranscoder
def test_update_pipeline_input_and_role():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    role = create_role_name("nonexistingrole")
    newrole = create_role_name("newrole")
    pipeline = client.create_pipeline(
        Name="testpipeline",
        InputBucket="inputtest",
        OutputBucket="outputtest",
        Role=role,
    )["Pipeline"]
    response = client.update_pipeline(
        Id=pipeline["Id"], InputBucket="inputtest2", Role=newrole
    )

    response["ResponseMetadata"]["HTTPStatusCode"].should.equal(200)
    response.should.have.key("Pipeline")

    pipeline = response["Pipeline"]
    pipeline.should.have.key("Id")
    pipeline.should.have.key("Name").being.equal("testpipeline")
    pipeline.should.have.key("InputBucket").being.equal("inputtest2")
    pipeline.should.have.key("Role").being.equal(newrole)


@mock_elastictranscoder
def test_update_pipeline_with_invalid_id():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    with pytest.raises(ClientError) as ex:
        client.update_pipeline(Id="unknown-pipeline")
    err = ex.value.response["Error"]
    err["Code"].should.equal("ValidationException")
    err["Message"].should.equal(
        "1 validation error detected: Value 'unknown-pipeline' at 'id' failed to satisfy constraint: Member must satisfy regular expression pattern: ^\\d{13}-\\w{6}$"
    )


@mock_elastictranscoder
def test_update_nonexisting_pipeline():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    pipeline_id = "0000000000000-abcdef"
    with pytest.raises(ClientError) as ex:
        client.read_pipeline(Id=pipeline_id)
    err = ex.value.response["Error"]
    err["Code"].should.equal("ResourceNotFoundException")
    err["Message"].should.equal(
        f"The specified pipeline was not found: account={ACCOUNT_ID}, pipelineId={pipeline_id}."
    )


@mock_elastictranscoder
def test_delete_pipeline():
    region = "us-east-1"
    client = boto3.client("elastictranscoder", region_name=region)

    role = create_role_name("nonexistingrole")
    pipeline = client.create_pipeline(
        Name="testpipeline",
        InputBucket="inputtest",
        OutputBucket="outputtest",
        Role=role,
    )["Pipeline"]
    client.delete_pipeline(Id=pipeline["Id"])

    response = client.list_pipelines()
    response["ResponseMetadata"]["HTTPStatusCode"].should.equal(200)
    response.should.have.key("Pipelines").being.length_of(0)


def create_role_name(name):
    return f"arn:aws:iam::{ACCOUNT_ID}:role/{name}"



import sure  # noqa # pylint: disable=unused-import

from moto.ses.utils import is_valid_address


def test_is_valid_address():
    valid, msg = is_valid_address("test@example.com")
    valid.should.equal(True)
    msg.should.equal(None)

    valid, msg = is_valid_address("test@")
    valid.should.equal(False)
    msg.should.be.a(str)

    valid, msg = is_valid_address("test")
    valid.should.equal(False)
    msg.should.be.a(str)



import boto3
import sure  # noqa # pylint: disable=unused-import
from moto import mock_codebuild
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID
from botocore.exceptions import ClientError, ParamValidationError
from uuid import uuid1
import pytest


@mock_codebuild
def test_codebuild_create_project_s3_artifacts():
    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    # output artifacts
    artifacts = dict()
    artifacts["type"] = "S3"
    artifacts["location"] = "bucketname"

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"

    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    response = client.create_project(
        name=name,
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )

    response.should.have.key("project")
    response["project"].should.have.key("serviceRole")
    response["project"].should.have.key("name").equals(name)

    response["project"]["environment"].should.equal(
        {
            "computeType": "BUILD_GENERAL1_SMALL",
            "image": "contents_not_validated",
            "type": "LINUX_CONTAINER",
        }
    )

    response["project"]["source"].should.equal(
        {"location": "bucketname/path/file.zip", "type": "S3"}
    )

    response["project"]["artifacts"].should.equal(
        {"location": "bucketname", "type": "S3"}
    )


@mock_codebuild
def test_codebuild_create_project_no_artifacts():
    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    # output artifacts
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    response = client.create_project(
        name=name,
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )

    response.should.have.key("project")
    response["project"].should.have.key("serviceRole")
    response["project"].should.have.key("name").equals(name)

    response["project"]["environment"].should.equal(
        {
            "computeType": "BUILD_GENERAL1_SMALL",
            "image": "contents_not_validated",
            "type": "LINUX_CONTAINER",
        }
    )

    response["project"]["source"].should.equal(
        {"location": "bucketname/path/file.zip", "type": "S3"}
    )

    response["project"]["artifacts"].should.equal({"type": "NO_ARTIFACTS"})


@mock_codebuild
def test_codebuild_create_project_with_invalid_name():
    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "!some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    # output artifacts
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    with pytest.raises(client.exceptions.from_code("InvalidInputException")) as err:
        client.create_project(
            name=name,
            source=source,
            artifacts=artifacts,
            environment=environment,
            serviceRole=service_role,
        )
    err.value.response["Error"]["Code"].should.equal("InvalidInputException")


@mock_codebuild
def test_codebuild_create_project_with_invalid_name_length():
    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project_" * 12
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    # output artifacts
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    with pytest.raises(client.exceptions.from_code("InvalidInputException")) as err:
        client.create_project(
            name=name,
            source=source,
            artifacts=artifacts,
            environment=environment,
            serviceRole=service_role,
        )
    err.value.response["Error"]["Code"].should.equal("InvalidInputException")


@mock_codebuild
def test_codebuild_create_project_when_exists():
    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    client.create_project(
        name=name,
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )

    with pytest.raises(ClientError) as err:
        client.create_project(
            name=name,
            source=source,
            artifacts=artifacts,
            environment=environment,
            serviceRole=service_role,
        )
    err.value.response["Error"]["Code"].should.equal("ResourceAlreadyExistsException")


@mock_codebuild
def test_codebuild_list_projects():
    client = boto3.client("codebuild", region_name="eu-central-1")

    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    # output artifacts
    artifacts = dict()
    artifacts["type"] = "S3"
    artifacts["location"] = "bucketname"

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    client.create_project(
        name="project1",
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )
    client.create_project(
        name="project2",
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )

    projects = client.list_projects()

    projects["projects"].should.equal(["project1", "project2"])


@mock_codebuild
def test_codebuild_list_builds_for_project_no_history():
    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    client.create_project(
        name=name,
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )
    history = client.list_builds_for_project(projectName=name)

    # no build history if it's never started
    history["ids"].should.equal([])


@mock_codebuild
def test_codebuild_list_builds_for_project_with_history():
    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    client.create_project(
        name=name,
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )
    client.start_build(projectName=name)
    response = client.list_builds_for_project(projectName=name)

    response["ids"].should.have.length_of(1)


# project never started
@mock_codebuild
def test_codebuild_get_batch_builds_for_project_no_history():
    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    client.create_project(
        name=name,
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )

    response = client.list_builds_for_project(projectName=name)
    response["ids"].should.equal([])

    with pytest.raises(ParamValidationError) as err:
        client.batch_get_builds(ids=response["ids"])
    err.typename.should.equal("ParamValidationError")


@mock_codebuild
def test_codebuild_start_build_no_project():

    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"

    with pytest.raises(client.exceptions.from_code("ResourceNotFoundException")) as err:
        client.start_build(projectName=name)
    err.value.response["Error"]["Code"].should.equal("ResourceNotFoundException")


@mock_codebuild
def test_codebuild_start_build_no_overrides():

    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    client.create_project(
        name=name,
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )
    response = client.start_build(projectName=name)

    response.should.have.key("build")
    response["build"]["sourceVersion"].should.equal("refs/heads/main")


@mock_codebuild
def test_codebuild_start_build_multiple_times():
    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    client.create_project(
        name=name,
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )

    client.start_build(projectName=name)
    client.start_build(projectName=name)
    client.start_build(projectName=name)

    len(client.list_builds()["ids"]).should.equal(3)


@mock_codebuild
def test_codebuild_start_build_with_overrides():

    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    branch_override = "fix/testing"
    artifacts_override = {"type": "NO_ARTIFACTS"}

    client.create_project(
        name=name,
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )
    response = client.start_build(
        projectName=name,
        sourceVersion=branch_override,
        artifactsOverride=artifacts_override,
    )

    response.should.have.key("build")
    response["build"]["sourceVersion"].should.equal("fix/testing")


@mock_codebuild
def test_codebuild_batch_get_builds_1_project():
    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    client.create_project(
        name=name,
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )
    client.start_build(projectName=name)

    history = client.list_builds_for_project(projectName=name)
    response = client.batch_get_builds(ids=history["ids"])

    response.should.have.key("builds").length_of(1)
    response["builds"][0]["currentPhase"].should.equal("COMPLETED")
    response["builds"][0]["buildNumber"].should.be.a(int)
    response["builds"][0].should.have.key("phases")
    len(response["builds"][0]["phases"]).should.equal(11)


@mock_codebuild
def test_codebuild_batch_get_builds_2_projects():
    client = boto3.client("codebuild", region_name="eu-central-1")

    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    client.create_project(
        name="project-1",
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )
    client.start_build(projectName="project-1")

    client.create_project(
        name="project-2",
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )
    client.start_build(projectName="project-2")

    response = client.list_builds()
    response["ids"].should.have.length_of(2)

    "project-1".should.be.within(response["ids"][0])
    "project-2".should.be.within(response["ids"][1])

    metadata = client.batch_get_builds(ids=response["ids"])["builds"]
    metadata.should.have.length_of(2)

    "project-1".should.be.within(metadata[0]["id"])
    "project-2".should.be.within(metadata[1]["id"])


@mock_codebuild
def test_codebuild_batch_get_builds_invalid_build_id():
    client = boto3.client("codebuild", region_name="eu-central-1")

    with pytest.raises(client.exceptions.InvalidInputException) as err:
        client.batch_get_builds(ids=[f"some_project{uuid1()}"])
    err.value.response["Error"]["Code"].should.equal("InvalidInputException")


@mock_codebuild
def test_codebuild_batch_get_builds_empty_build_id():
    client = boto3.client("codebuild", region_name="eu-central-1")

    with pytest.raises(ParamValidationError) as err:
        client.batch_get_builds(ids=[])
    err.typename.should.equal("ParamValidationError")


@mock_codebuild
def test_codebuild_delete_project():
    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    client.create_project(
        name=name,
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )
    client.start_build(projectName=name)

    response = client.list_builds_for_project(projectName=name)
    response["ids"].should.have.length_of(1)

    client.delete_project(name=name)

    with pytest.raises(ClientError) as err:
        client.list_builds_for_project(projectName=name)
    err.value.response["Error"]["Code"].should.equal("ResourceNotFoundException")


@mock_codebuild
def test_codebuild_stop_build():
    client = boto3.client("codebuild", region_name="eu-central-1")

    name = "some_project"
    source = dict()
    source["type"] = "S3"
    # repository location for S3
    source["location"] = "bucketname/path/file.zip"
    artifacts = {"type": "NO_ARTIFACTS"}

    environment = dict()
    environment["type"] = "LINUX_CONTAINER"
    environment["image"] = "contents_not_validated"
    environment["computeType"] = "BUILD_GENERAL1_SMALL"
    service_role = (
        f"arn:aws:iam::{ACCOUNT_ID}:role/service-role/my-codebuild-service-role"
    )

    client.create_project(
        name=name,
        source=source,
        artifacts=artifacts,
        environment=environment,
        serviceRole=service_role,
    )
    client.start_build(projectName=name)

    builds = client.list_builds()

    response = client.stop_build(id=builds["ids"][0])
    response["build"]["buildStatus"].should.equal("STOPPED")


@mock_codebuild
def test_codebuild_stop_build_no_build():
    client = boto3.client("codebuild", region_name="eu-central-1")

    with pytest.raises(client.exceptions.ResourceNotFoundException) as err:
        client.stop_build(id=f"some_project:{uuid1()}")
    err.value.response["Error"]["Code"].should.equal("ResourceNotFoundException")


@mock_codebuild
def test_codebuild_stop_build_bad_uid():
    client = boto3.client("codebuild", region_name="eu-central-1")

    with pytest.raises(client.exceptions.InvalidInputException) as err:
        client.stop_build(id=f"some_project{uuid1()}")
    err.value.response["Error"]["Code"].should.equal("InvalidInputException")



import boto3
import sure  # noqa # pylint: disable=unused-import
import pytest

from botocore.exceptions import ClientError
from moto import mock_ec2, mock_route53


@mock_ec2
@mock_route53
def test_hosted_zone_private_zone_preserved():
    # Create mock VPC so we can get a VPC ID
    region = "us-east-1"
    ec2c = boto3.client("ec2", region_name=region)
    vpc_id = ec2c.create_vpc(CidrBlock="10.1.0.0/16").get("Vpc").get("VpcId")

    # Create hosted_zone as a Private VPC Hosted Zone
    conn = boto3.client("route53", region_name=region)
    new_zone = conn.create_hosted_zone(
        Name="testdns.aws.com.",
        CallerReference=str(hash("foo")),
        HostedZoneConfig=dict(PrivateZone=True, Comment="Test"),
        VPC={"VPCRegion": region, "VPCId": vpc_id},
    )

    zone_id = new_zone["HostedZone"]["Id"].split("/")[-1]
    hosted_zone = conn.get_hosted_zone(Id=zone_id)
    hosted_zone["HostedZone"]["Config"]["PrivateZone"].should.equal(True)
    hosted_zone.should.have.key("VPCs")
    hosted_zone["VPCs"].should.have.length_of(1)
    hosted_zone["VPCs"][0].should.have.key("VPCId")
    hosted_zone["VPCs"][0].should.have.key("VPCRegion")
    hosted_zone["VPCs"][0]["VPCId"].should.be.equal(vpc_id)
    hosted_zone["VPCs"][0]["VPCRegion"].should.be.equal(region)

    hosted_zones = conn.list_hosted_zones()
    hosted_zones["HostedZones"][0]["Config"]["PrivateZone"].should.equal(True)

    hosted_zones = conn.list_hosted_zones_by_name(DNSName="testdns.aws.com.")
    hosted_zones["HostedZones"].should.have.length_of(1)
    hosted_zones["HostedZones"][0]["Config"]["PrivateZone"].should.equal(True)

    # create_hosted_zone statements with  PrivateZone=True,
    # but without a _valid_ vpc-id should NOT fail.
    zone2_name = "testdns2.aws.com."
    no_vpc_zone = conn.create_hosted_zone(
        Name=zone2_name,
        CallerReference=str(hash("foo")),
        HostedZoneConfig=dict(PrivateZone=True, Comment="Test without VPC"),
    )

    zone_id = no_vpc_zone["HostedZone"]["Id"].split("/")[-1]
    hosted_zone = conn.get_hosted_zone(Id=zone_id)
    hosted_zone["HostedZone"]["Config"]["PrivateZone"].should.equal(True)
    hosted_zone.should.have.key("VPCs")
    hosted_zone["VPCs"].should.have.length_of(0)

    hosted_zones = conn.list_hosted_zones()
    hosted_zones["HostedZones"].should.have.length_of(2)
    hosted_zones["HostedZones"][0]["Config"]["PrivateZone"].should.equal(True)
    hosted_zones["HostedZones"][1]["Config"]["PrivateZone"].should.equal(True)

    hosted_zones = conn.list_hosted_zones_by_name(DNSName=zone2_name)
    hosted_zones["HostedZones"].should.have.length_of(1)
    hosted_zones["HostedZones"][0]["Config"]["PrivateZone"].should.equal(True)
    hosted_zones["HostedZones"][0]["Name"].should.equal(zone2_name)


@mock_ec2
@mock_route53
def test_list_hosted_zones_by_vpc_with_multiple_vpcs():
    # Create mock VPC so we can get a VPC ID
    ec2c = boto3.client("ec2", region_name="us-east-1")
    vpc_id = ec2c.create_vpc(CidrBlock="10.1.0.0/16").get("Vpc").get("VpcId")
    region = "us-east-1"

    # Create 3 Zones associate with the VPC.
    zones = {}
    conn = boto3.client("route53", region_name=region)
    for zone in ["a", "b", "c"]:
        zone_name = f"test.{zone}.com."
        zones[zone] = conn.create_hosted_zone(
            Name=zone_name,
            CallerReference=str(hash("foo")),
            HostedZoneConfig=dict(PrivateZone=True, Comment=f"test {zone} com"),
            VPC={"VPCRegion": region, "VPCId": vpc_id},
        )

    # List the zones associated with this vpc
    response = conn.list_hosted_zones_by_vpc(VPCId=vpc_id, VPCRegion=region)
    response.should.have.key("ResponseMetadata")
    response.should.have.key("HostedZoneSummaries")
    response["HostedZoneSummaries"].should.have.length_of(3)

    # Loop through all zone summaries and verify they match what was created
    for summary in response["HostedZoneSummaries"]:
        # use the zone name as the index
        index = summary["Name"].split(".")[1]
        zone_id = zones[index]["HostedZone"]["Id"].split("/")[2]
        summary.should.have.key("HostedZoneId")
        summary["HostedZoneId"].should.equal(zone_id)
        summary.should.have.key("Name")
        summary["Name"].should.equal(zones[index]["HostedZone"]["Name"])


@mock_ec2
@mock_route53
def test_list_hosted_zones_by_vpc():
    # Create mock VPC so we can get a VPC ID
    ec2c = boto3.client("ec2", region_name="us-east-1")
    vpc_id = ec2c.create_vpc(CidrBlock="10.1.0.0/16").get("Vpc").get("VpcId")
    region = "us-east-1"

    conn = boto3.client("route53", region_name=region)
    zone_b = conn.create_hosted_zone(
        Name="test.b.com.",
        CallerReference=str(hash("foo")),
        HostedZoneConfig=dict(PrivateZone=True, Comment="test com"),
        VPC={"VPCRegion": region, "VPCId": vpc_id},
    )
    zone_id = zone_b["HostedZone"]["Id"].split("/")[2]

    response = conn.list_hosted_zones_by_vpc(VPCId=vpc_id, VPCRegion=region)
    response.should.have.key("ResponseMetadata")
    response.should.have.key("HostedZoneSummaries")
    response["HostedZoneSummaries"].should.have.length_of(1)
    response["HostedZoneSummaries"][0].should.have.key("HostedZoneId")
    retured_zone = response["HostedZoneSummaries"][0]
    retured_zone["HostedZoneId"].should.equal(zone_id)
    retured_zone["Name"].should.equal(zone_b["HostedZone"]["Name"])


@mock_ec2
@mock_route53
def test_route53_associate_vpc():
    ec2c = boto3.client("ec2", region_name="us-east-1")
    vpc_id = ec2c.create_vpc(CidrBlock="10.1.0.0/16")["Vpc"]["VpcId"]
    conn = boto3.client("route53", region_name="us-east-1")
    zone = conn.create_hosted_zone(
        Name="test.b.com.",
        CallerReference=str(hash("foo")),
        HostedZoneConfig=dict(PrivateZone=True, Comment=""),
    )
    zone_id = zone["HostedZone"]["Id"].split("/")[2]

    resp = conn.associate_vpc_with_hosted_zone(
        HostedZoneId=zone_id,
        VPC={"VPCId": vpc_id, "VPCRegion": "us-east-1"},
        Comment="yolo",
    )
    resp.should.have.key("ChangeInfo")
    resp["ChangeInfo"].should.have.key("Comment").equals("yolo")


@mock_ec2
@mock_route53
def test_route53_associate_vpc_with_public_Zone():
    ec2c = boto3.client("ec2", region_name="us-east-1")
    vpc_id = ec2c.create_vpc(CidrBlock="10.1.0.0/16")["Vpc"]["VpcId"]
    conn = boto3.client("route53", region_name="us-east-1")
    zone = conn.create_hosted_zone(
        Name="test.b.com.",
        CallerReference=str(hash("foo")),
    )
    zone_id = zone["HostedZone"]["Id"].split("/")[2]

    with pytest.raises(ClientError) as exc:
        conn.associate_vpc_with_hosted_zone(
            HostedZoneId=zone_id,
            VPC={"VPCId": vpc_id, "VPCRegion": "us-east-1"},
            Comment="yolo",
        )
    err = exc.value.response["Error"]
    err["Code"].should.equal("PublicZoneVPCAssociation")
    err["Message"].should.equal(
        "You're trying to associate a VPC with a public hosted zone. Amazon Route 53 doesn't support associating a VPC with a public hosted zone."
    )


@mock_ec2
@mock_route53
def test_route53_associate_and_disassociate_vpc():
    ec2c = boto3.client("ec2", region_name="us-east-1")
    vpc_id1 = ec2c.create_vpc(CidrBlock="10.1.0.0/16").get("Vpc").get("VpcId")
    vpc_id2 = ec2c.create_vpc(CidrBlock="10.1.0.1/16").get("Vpc").get("VpcId")
    region = "us-east-1"

    conn = boto3.client("route53", region_name=region)
    zone = conn.create_hosted_zone(
        Name="test.b.com.",
        CallerReference=str(hash("foo")),
        HostedZoneConfig=dict(PrivateZone=True, Comment="test com"),
        VPC={"VPCRegion": region, "VPCId": vpc_id1},
    )
    zone_id = zone["HostedZone"]["Id"].split("/")[2]

    conn.associate_vpc_with_hosted_zone(
        HostedZoneId=zone_id,
        VPC={"VPCId": vpc_id2, "VPCRegion": region},
    )

    zone_vpcs = conn.get_hosted_zone(Id=zone_id)["VPCs"]
    zone_vpcs.should.have.length_of(2)
    zone_vpcs.should.contain({"VPCRegion": region, "VPCId": vpc_id1})
    zone_vpcs.should.contain({"VPCRegion": region, "VPCId": vpc_id2})

    conn.disassociate_vpc_from_hosted_zone(HostedZoneId=zone_id, VPC={"VPCId": vpc_id1})

    zone_vpcs = conn.get_hosted_zone(Id=zone_id)["VPCs"]
    zone_vpcs.should.have.length_of(1)
    zone_vpcs.should.contain({"VPCRegion": region, "VPCId": vpc_id2})


@mock_ec2
@mock_route53
def test_route53_disassociate_last_vpc():
    ec2c = boto3.client("ec2", region_name="us-east-1")
    vpc_id = ec2c.create_vpc(CidrBlock="10.1.0.0/16")["Vpc"]["VpcId"]
    conn = boto3.client("route53", region_name="us-east-1")
    zone = conn.create_hosted_zone(
        Name="test.b.com.",
        CallerReference=str(hash("foo")),
        HostedZoneConfig=dict(PrivateZone=True, Comment="test com"),
        VPC={"VPCRegion": "us-east-1", "VPCId": vpc_id},
    )
    zone_id = zone["HostedZone"]["Id"].split("/")[2]

    with pytest.raises(ClientError) as exc:
        conn.disassociate_vpc_from_hosted_zone(
            HostedZoneId=zone_id, VPC={"VPCId": vpc_id}
        )
    err = exc.value.response["Error"]
    err["Code"].should.equal("LastVPCAssociation")
    err["Message"].should.equal(
        "The VPC that you're trying to disassociate from the private hosted zone is the last VPC that is associated with the hosted zone. Amazon Route 53 doesn't support disassociating the last VPC from a hosted zone."
    )



from moto.core import BaseBackend, BackendDict


class InstanceMetadataBackend(BaseBackend):
    pass


instance_metadata_backends = BackendDict(
    InstanceMetadataBackend,
    "instance_metadata",
    use_boto3_regions=False,
    additional_regions=["global"],
)



import sure  # noqa # pylint: disable=unused-import

import moto.server as server
from moto import mock_mediastore

"""
Test the different server responses
"""


@mock_mediastore
def test_mediastore_lists_containers():
    backend = server.create_backend_app("mediastore")
    test_client = backend.test_client()

    res = test_client.get(
        "/", headers={"X-Amz-Target": "MediaStore_20170901.ListContainers"}
    )

    result = res.data.decode("utf-8")
    result.should.contain('"Containers": []')



"""Unit tests for servicediscovery-supported APIs."""
import boto3
import pytest
import sure  # noqa # pylint: disable=unused-import

from botocore.exceptions import ClientError
from moto import mock_servicediscovery

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_servicediscovery
def test_list_operations_initial():
    client = boto3.client("servicediscovery", region_name="eu-west-1")
    resp = client.list_operations()

    resp.should.have.key("Operations").equals([])


@mock_servicediscovery
def test_list_operations():
    client = boto3.client("servicediscovery", region_name="eu-west-2")

    resp = client.create_http_namespace(Name="n/a")
    resp.should.have.key("OperationId")
    op_id = resp["OperationId"]

    resp = client.list_operations()
    resp.should.have.key("Operations").length_of(1)
    resp["Operations"].should.equal([{"Id": op_id, "Status": "SUCCESS"}])


@mock_servicediscovery
def test_get_create_http_namespace_operation():
    client = boto3.client("servicediscovery", region_name="eu-west-1")
    resp = client.create_http_namespace(Name="mynamespace")

    resp["OperationId"].should.match("[a-z0-9]{32}-[a-z0-9]{8}")

    operation_id = resp["OperationId"]

    resp = client.get_operation(OperationId=operation_id)

    resp.should.have.key("Operation")
    operation = resp["Operation"]
    operation.should.have.key("Id").equals(operation_id)
    operation.should.have.key("Type").equals("CREATE_NAMESPACE")
    operation.should.have.key("Status").equals("SUCCESS")
    operation.should.have.key("CreateDate")
    operation.should.have.key("UpdateDate")
    operation.should.have.key("Targets")

    targets = operation["Targets"]
    targets.should.have.key("NAMESPACE")

    namespaces = client.list_namespaces()["Namespaces"]
    [ns["Id"] for ns in namespaces].should.contain(targets["NAMESPACE"])


@mock_servicediscovery
def test_get_private_dns_namespace_operation():
    client = boto3.client("servicediscovery", region_name="eu-west-1")
    resp = client.create_private_dns_namespace(Name="dns_ns", Vpc="vpc_id")

    resp["OperationId"].should.match("[a-z0-9]{32}-[a-z0-9]{8}")

    operation_id = resp["OperationId"]

    resp = client.get_operation(OperationId=operation_id)

    resp.should.have.key("Operation")
    operation = resp["Operation"]
    operation.should.have.key("Id").equals(operation_id)
    operation.should.have.key("Type").equals("CREATE_NAMESPACE")
    operation.should.have.key("Status").equals("SUCCESS")
    operation.should.have.key("CreateDate")
    operation.should.have.key("UpdateDate")
    operation.should.have.key("Targets")


@mock_servicediscovery
def test_get_public_dns_namespace_operation():
    client = boto3.client("servicediscovery", region_name="eu-west-1")
    resp = client.create_public_dns_namespace(Name="dns_ns")

    resp["OperationId"].should.match("[a-z0-9]{32}-[a-z0-9]{8}")

    operation_id = resp["OperationId"]

    resp = client.get_operation(OperationId=operation_id)

    resp.should.have.key("Operation")
    operation = resp["Operation"]
    operation.should.have.key("Id").equals(operation_id)
    operation.should.have.key("Type").equals("CREATE_NAMESPACE")
    operation.should.have.key("Status").equals("SUCCESS")
    operation.should.have.key("CreateDate")
    operation.should.have.key("UpdateDate")
    operation.should.have.key("Targets")


@mock_servicediscovery
def test_get_update_service_operation():
    client = boto3.client("servicediscovery", region_name="eu-west-1")
    service_id = client.create_service(
        Name="my service", NamespaceId="ns_id", Description="first desc"
    )["Service"]["Id"]

    resp = client.update_service(Id=service_id, Service={"Description": "updated desc"})

    resp["OperationId"].should.match("[a-z0-9]{32}-[a-z0-9]{8}")

    operation_id = resp["OperationId"]

    resp = client.get_operation(OperationId=operation_id)

    resp.should.have.key("Operation")
    operation = resp["Operation"]
    operation.should.have.key("Id").equals(operation_id)
    operation.should.have.key("Type").equals("UPDATE_SERVICE")
    operation.should.have.key("Status").equals("SUCCESS")
    operation.should.have.key("CreateDate")
    operation.should.have.key("UpdateDate")
    operation.should.have.key("Targets")


@mock_servicediscovery
def test_get_unknown_operation():
    client = boto3.client("servicediscovery", region_name="eu-west-1")

    with pytest.raises(ClientError) as exc:
        client.get_operation(OperationId="unknown")
    err = exc.value.response["Error"]
    err["Code"].should.equal("OperationNotFound")



import boto3

from datetime import timedelta, datetime
from moto import mock_appsync
from moto.core.utils import unix_time


@mock_appsync
def test_create_api_key_simple():
    client = boto3.client("appsync", region_name="eu-west-1")

    api_id = client.create_graphql_api(name="api1", authenticationType="API_KEY")[
        "graphqlApi"
    ]["apiId"]
    resp = client.create_api_key(apiId=api_id)

    resp.should.have.key("apiKey")
    api_key = resp["apiKey"]

    api_key.should.have.key("id")
    api_key.shouldnt.have.key("description")
    api_key.should.have.key("expires")
    api_key.should.have.key("deletes")


@mock_appsync
def test_create_api_key():
    client = boto3.client("appsync", region_name="ap-southeast-1")
    tomorrow = datetime.now() + timedelta(days=1)
    tomorrow_in_secs = int(unix_time(tomorrow))

    api_id = client.create_graphql_api(name="api1", authenticationType="API_KEY")[
        "graphqlApi"
    ]["apiId"]
    resp = client.create_api_key(
        apiId=api_id, description="my first api key", expires=tomorrow_in_secs
    )

    resp.should.have.key("apiKey")
    api_key = resp["apiKey"]

    api_key.should.have.key("id")
    api_key.should.have.key("description").equals("my first api key")
    api_key.should.have.key("expires").equals(tomorrow_in_secs)
    api_key.should.have.key("deletes").equals(tomorrow_in_secs)


@mock_appsync
def test_delete_api_key():
    client = boto3.client("appsync", region_name="us-east-1")

    api_id = client.create_graphql_api(name="api1", authenticationType="API_KEY")[
        "graphqlApi"
    ]["apiId"]
    api_key_id = client.create_api_key(apiId=api_id)["apiKey"]["id"]

    client.delete_api_key(apiId=api_id, id=api_key_id)

    resp = client.list_api_keys(apiId=api_id)
    resp.should.have.key("apiKeys").length_of(0)


@mock_appsync
def test_list_api_keys_unknown_api():
    client = boto3.client("appsync", region_name="ap-southeast-1")
    resp = client.list_api_keys(apiId="unknown")
    resp.should.have.key("apiKeys").equals([])


@mock_appsync
def test_list_api_keys_empty():
    client = boto3.client("appsync", region_name="ap-southeast-1")
    api_id = client.create_graphql_api(name="api1", authenticationType="API_KEY")[
        "graphqlApi"
    ]["apiId"]

    resp = client.list_api_keys(apiId=api_id)
    resp.should.have.key("apiKeys").equals([])


@mock_appsync
def test_list_api_keys():
    client = boto3.client("appsync", region_name="ap-southeast-1")
    api_id = client.create_graphql_api(name="api1", authenticationType="API_KEY")[
        "graphqlApi"
    ]["apiId"]
    client.create_api_key(apiId=api_id)
    client.create_api_key(apiId=api_id, description="my first api key")
    resp = client.list_api_keys(apiId=api_id)
    resp.should.have.key("apiKeys").length_of(2)


@mock_appsync
def test_update_api_key():
    client = boto3.client("appsync", region_name="eu-west-1")

    api_id = client.create_graphql_api(name="api1", authenticationType="API_KEY")[
        "graphqlApi"
    ]["apiId"]

    original = client.create_api_key(apiId=api_id, description="my first api key")[
        "apiKey"
    ]

    updated = client.update_api_key(
        apiId=api_id, id=original["id"], description="my second api key"
    )["apiKey"]

    updated.should.have.key("id").equals(original["id"])
    updated.should.have.key("description").equals("my second api key")
    updated.should.have.key("expires").equals(original["expires"])
    updated.should.have.key("deletes").equals(original["deletes"])



"""Unit tests for glue-supported APIs."""
from random import randint
from uuid import uuid4

import boto3
import pytest
import sure  # noqa # pylint: disable=unused-import
from botocore.exceptions import ParamValidationError
from botocore.client import ClientError

from moto import mock_glue


@mock_glue
def test_create_job():
    client = create_glue_client()
    job_name = str(uuid4())
    response = client.create_job(
        Name=job_name, Role="test_role", Command=dict(Name="test_command")
    )
    assert response["Name"] == job_name


@mock_glue
def test_create_job_default_argument_not_provided():
    client = create_glue_client()
    with pytest.raises(ParamValidationError) as exc:
        client.create_job(Role="test_role", Command=dict(Name="test_command"))

    exc.value.kwargs["report"].should.equal(
        'Missing required parameter in input: "Name"'
    )


@mock_glue
def test_list_jobs():
    client = create_glue_client()
    expected_jobs = randint(1, 15)
    create_test_jobs(client, expected_jobs)
    response = client.list_jobs()
    response["JobNames"].should.have.length_of(expected_jobs)
    response.shouldnt.have.key("NextToken")


@mock_glue
def test_get_job_not_exists():
    client = create_glue_client()
    name = "my_job_name"

    with pytest.raises(ClientError) as exc:
        client.get_job(JobName=name)

    exc.value.response["Error"]["Code"].should.equal("EntityNotFoundException")
    exc.value.response["Error"]["Message"].should.match("Job my_job_name not found")


@mock_glue
def test_get_job_exists():
    client = create_glue_client()
    job_attributes = {
        "Description": "test_description",
        "LogUri": "test_log/",
        "Role": "test_role",
        "ExecutionProperty": {"MaxConcurrentRuns": 123},
        "Command": {
            "Name": "test_command",
            "ScriptLocation": "test_s3_path",
            "PythonVersion": "3.6",
        },
        "DefaultArguments": {"string": "string"},
        "NonOverridableArguments": {"string": "string"},
        "Connections": {
            "Connections": [
                "string",
            ]
        },
        "MaxRetries": 123,
        "AllocatedCapacity": 123,
        "Timeout": 123,
        "MaxCapacity": 123.0,
        "WorkerType": "G.2X",
        "NumberOfWorkers": 123,
        "SecurityConfiguration": "test_config",
        "NotificationProperty": {"NotifyDelayAfter": 123},
        "GlueVersion": "string",
    }
    job_name = create_test_job_w_all_attributes(client, **job_attributes)
    job = client.get_job(JobName=job_name)["Job"]
    job.should.have.key("Name").equals(job_name)
    job.should.have.key("Description")
    job.should.have.key("LogUri")
    job.should.have.key("Role")
    job.should.have.key("ExecutionProperty").equals({"MaxConcurrentRuns": 123})
    job.should.have.key("CreatedOn")
    job.should.have.key("LastModifiedOn")
    job.should.have.key("ExecutionProperty")
    job.should.have.key("Command")
    job.should.have.key("DefaultArguments")
    job.should.have.key("NonOverridableArguments")
    job.should.have.key("Connections")
    job.should.have.key("MaxRetries")
    job.should.have.key("AllocatedCapacity")
    job.should.have.key("Timeout")
    job.should.have.key("MaxCapacity")
    job.should.have.key("WorkerType")
    job.should.have.key("NumberOfWorkers")
    job.should.have.key("SecurityConfiguration")
    job.should.have.key("NotificationProperty")
    job.should.have.key("GlueVersion")


@mock_glue
def test_list_jobs_with_max_results():
    client = create_glue_client()
    create_test_jobs(client, 4)
    response = client.list_jobs(MaxResults=2)
    response["JobNames"].should.have.length_of(2)
    response.should.have.key("NextToken")


@mock_glue
def test_list_jobs_from_next_token():
    client = create_glue_client()
    create_test_jobs(client, 10)
    first_response = client.list_jobs(MaxResults=3)
    response = client.list_jobs(NextToken=first_response["NextToken"])
    response["JobNames"].should.have.length_of(7)


@mock_glue
def test_list_jobs_with_max_results_greater_than_actual_results():
    client = create_glue_client()
    create_test_jobs(client, 4)
    response = client.list_jobs(MaxResults=10)
    response["JobNames"].should.have.length_of(4)


@mock_glue
def test_list_jobs_with_tags():
    client = create_glue_client()
    create_test_job(client)
    create_test_job(client, {"string": "string"})
    response = client.list_jobs(Tags={"string": "string"})
    response["JobNames"].should.have.length_of(1)


@mock_glue
def test_list_jobs_after_tagging():
    client = create_glue_client()
    job_name = create_test_job(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:job/{job_name}"

    client.tag_resource(ResourceArn=resource_arn, TagsToAdd={"key1": "value1"})

    response = client.list_jobs(Tags={"key1": "value1"})
    response["JobNames"].should.have.length_of(1)


@mock_glue
def test_list_jobs_after_removing_tag():
    client = create_glue_client()
    job_name = create_test_job(client, {"key1": "value1"})
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:job/{job_name}"

    client.untag_resource(ResourceArn=resource_arn, TagsToRemove=["key1"])

    response = client.list_jobs(Tags={"key1": "value1"})
    response["JobNames"].should.have.length_of(0)


@mock_glue
def test_list_jobs_next_token_logic_does_not_create_infinite_loop():
    client = create_glue_client()
    create_test_jobs(client, 4)
    first_response = client.list_jobs(MaxResults=1)
    next_token = first_response["NextToken"]
    while next_token:
        response = client.list_jobs(NextToken=next_token)
        next_token = response.get("NextToken")
    assert not next_token


def create_glue_client():
    return boto3.client("glue", region_name="us-east-1")


def create_test_job(client, tags=None):
    job_name = str(uuid4())
    client.create_job(
        Name=job_name,
        Role="test_role",
        Command=dict(Name="test_command"),
        Tags=tags or {},
    )
    return job_name


def create_test_job_w_all_attributes(client, **job_attributes):
    job_name = str(uuid4())
    client.create_job(Name=job_name, **job_attributes)
    return job_name


def create_test_jobs(client, number_of_jobs):
    for _ in range(number_of_jobs):
        create_test_job(client)


def create_test_crawler(client, tags=None):
    crawler_name = str(uuid4())
    client.create_crawler(
        Name=crawler_name,
        Role="test_role",
        Targets={"S3Targets": [{"Path": "s3://tests3target"}]},
        Tags=tags or {},
    )
    return crawler_name


def create_test_crawlers(client, number_of_crawlers):
    for _ in range(number_of_crawlers):
        create_test_crawler(client)


@mock_glue
def test_list_crawlers_with_max_results():
    client = create_glue_client()
    create_test_crawlers(client, 4)
    response = client.list_crawlers(MaxResults=2)
    response["CrawlerNames"].should.have.length_of(2)
    response.should.have.key("NextToken")


@mock_glue
def test_list_crawlers_from_next_token():
    client = create_glue_client()
    create_test_crawlers(client, 10)
    first_response = client.list_crawlers(MaxResults=3)
    response = client.list_crawlers(NextToken=first_response["NextToken"])
    response["CrawlerNames"].should.have.length_of(7)


@mock_glue
def test_list_crawlers_with_max_results_greater_than_actual_results():
    client = create_glue_client()
    create_test_crawlers(client, 4)
    response = client.list_crawlers(MaxResults=10)
    response["CrawlerNames"].should.have.length_of(4)


@mock_glue
def test_list_crawlers_with_tags():
    client = create_glue_client()
    create_test_crawler(client)
    create_test_crawler(client, {"string": "string"})
    response = client.list_crawlers(Tags={"string": "string"})
    response["CrawlerNames"].should.have.length_of(1)


@mock_glue
def test_list_crawlers_after_tagging():
    client = create_glue_client()
    crawler_name = create_test_crawler(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:crawler/{crawler_name}"

    client.tag_resource(ResourceArn=resource_arn, TagsToAdd={"key1": "value1"})

    response = client.list_crawlers(Tags={"key1": "value1"})
    response["CrawlerNames"].should.have.length_of(1)


@mock_glue
def test_list_crawlers_after_removing_tag():
    client = create_glue_client()
    crawler_name = create_test_crawler(client, {"key1": "value1"})
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:crawler/{crawler_name}"

    client.untag_resource(ResourceArn=resource_arn, TagsToRemove=["key1"])

    response = client.list_crawlers(Tags={"key1": "value1"})
    response["CrawlerNames"].should.have.length_of(0)


@mock_glue
def test_list_crawlers_next_token_logic_does_not_create_infinite_loop():
    client = create_glue_client()
    create_test_crawlers(client, 4)
    first_response = client.list_crawlers(MaxResults=1)
    next_token = first_response["NextToken"]
    while next_token:
        response = client.list_crawlers(NextToken=next_token)
        next_token = response.get("NextToken")
    assert not next_token


@mock_glue
def test_get_tags_job():
    client = create_glue_client()
    job_name = create_test_job(client, {"key1": "value1", "key2": "value2"})
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:job/{job_name}"

    resp = client.get_tags(ResourceArn=resource_arn)

    resp.should.have.key("Tags").equals({"key1": "value1", "key2": "value2"})


@mock_glue
def test_get_tags_jobs_no_tags():
    client = create_glue_client()
    job_name = create_test_job(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:job/{job_name}"

    resp = client.get_tags(ResourceArn=resource_arn)

    resp.should.have.key("Tags").equals({})


@mock_glue
def test_tag_glue_job():
    client = create_glue_client()
    job_name = create_test_job(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:job/{job_name}"

    client.tag_resource(
        ResourceArn=resource_arn, TagsToAdd={"key1": "value1", "key2": "value2"}
    )

    resp = client.get_tags(ResourceArn=resource_arn)

    resp.should.have.key("Tags").equals({"key1": "value1", "key2": "value2"})


@mock_glue
def test_untag_glue_job():
    client = create_glue_client()
    job_name = create_test_job(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:job/{job_name}"

    client.tag_resource(
        ResourceArn=resource_arn,
        TagsToAdd={"key1": "value1", "key2": "value2", "key3": "value3"},
    )

    client.untag_resource(ResourceArn=resource_arn, TagsToRemove=["key2"])

    resp = client.get_tags(ResourceArn=resource_arn)

    resp.should.have.key("Tags").equals({"key1": "value1", "key3": "value3"})


@mock_glue
def test_get_tags_crawler():
    client = create_glue_client()
    crawler_name = create_test_crawler(client, {"key1": "value1", "key2": "value2"})
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:crawler/{crawler_name}"

    resp = client.get_tags(ResourceArn=resource_arn)

    resp.should.have.key("Tags").equals({"key1": "value1", "key2": "value2"})


@mock_glue
def test_get_tags_crawler_no_tags():
    client = create_glue_client()
    crawler_name = create_test_crawler(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:crawler/{crawler_name}"

    resp = client.get_tags(ResourceArn=resource_arn)

    resp.should.have.key("Tags").equals({})


@mock_glue
def test_tag_glue_crawler():
    client = create_glue_client()
    crawler_name = create_test_crawler(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:crawler/{crawler_name}"

    client.tag_resource(
        ResourceArn=resource_arn, TagsToAdd={"key1": "value1", "key2": "value2"}
    )

    resp = client.get_tags(ResourceArn=resource_arn)

    resp.should.have.key("Tags").equals({"key1": "value1", "key2": "value2"})


@mock_glue
def test_untag_glue_crawler():
    client = create_glue_client()
    crawler_name = create_test_crawler(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:crawler/{crawler_name}"

    client.tag_resource(
        ResourceArn=resource_arn,
        TagsToAdd={"key1": "value1", "key2": "value2", "key3": "value3"},
    )

    client.untag_resource(ResourceArn=resource_arn, TagsToRemove=["key2"])

    resp = client.get_tags(ResourceArn=resource_arn)

    resp.should.have.key("Tags").equals({"key1": "value1", "key3": "value3"})



import pytest

import sure  # noqa # pylint: disable=unused-import
import boto3
from botocore.exceptions import ClientError
from moto import mock_wafv2
from .test_helper_functions import CREATE_WEB_ACL_BODY, LIST_WEB_ACL_BODY
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID


@mock_wafv2
def test_create_web_acl():

    conn = boto3.client("wafv2", region_name="us-east-1")
    res = conn.create_web_acl(**CREATE_WEB_ACL_BODY("John", "REGIONAL"))
    web_acl = res["Summary"]
    assert web_acl.get("Name") == "John"
    assert web_acl.get("ARN").startswith(
        f"arn:aws:wafv2:us-east-1:{ACCOUNT_ID}:regional/webacl/John/"
    )
    # Duplicate name - should raise error
    with pytest.raises(ClientError) as ex:
        conn.create_web_acl(**CREATE_WEB_ACL_BODY("John", "REGIONAL"))
    err = ex.value.response["Error"]
    err["Message"].should.contain(
        "AWS WAF could not perform the operation because some resource in your request is a duplicate of an existing one."
    )
    err["Code"].should.equal("WafV2DuplicateItem")

    res = conn.create_web_acl(**CREATE_WEB_ACL_BODY("Carl", "CLOUDFRONT"))
    web_acl = res["Summary"]
    assert web_acl.get("ARN").startswith(
        f"arn:aws:wafv2:global:{ACCOUNT_ID}:global/webacl/Carl/"
    )


@mock_wafv2
def test_create_web_acl_with_all_arguments():
    client = boto3.client("wafv2", region_name="us-east-2")
    web_acl_id = client.create_web_acl(
        Name="test",
        Scope="CLOUDFRONT",
        DefaultAction={"Allow": {}},
        Description="test desc",
        VisibilityConfig={
            "SampledRequestsEnabled": False,
            "CloudWatchMetricsEnabled": False,
            "MetricName": "idk",
        },
        Rules=[
            {
                "Action": {"Allow": {}},
                "Name": "tf-acc-test-8205974093017792151-2",
                "Priority": 10,
                "Statement": {"GeoMatchStatement": {"CountryCodes": ["US", "NL"]}},
                "VisibilityConfig": {
                    "CloudWatchMetricsEnabled": False,
                    "MetricName": "tf-acc-test-8205974093017792151-2",
                    "SampledRequestsEnabled": False,
                },
            },
            {
                "Action": {"Count": {}},
                "Name": "tf-acc-test-8205974093017792151-1",
                "Priority": 5,
                "Statement": {
                    "SizeConstraintStatement": {
                        "ComparisonOperator": "LT",
                        "FieldToMatch": {"QueryString": {}},
                        "Size": 50,
                        "TextTransformations": [
                            {"Priority": 2, "Type": "CMD_LINE"},
                            {"Priority": 5, "Type": "NONE"},
                        ],
                    }
                },
                "VisibilityConfig": {
                    "CloudWatchMetricsEnabled": False,
                    "MetricName": "tf-acc-test-8205974093017792151-1",
                    "SampledRequestsEnabled": False,
                },
            },
        ],
    )["Summary"]["Id"]

    wacl = client.get_web_acl(Name="test", Scope="CLOUDFRONT", Id=web_acl_id)["WebACL"]
    wacl.should.have.key("Description").equals("test desc")
    wacl.should.have.key("DefaultAction").equals({"Allow": {}})
    wacl.should.have.key("VisibilityConfig").equals(
        {
            "SampledRequestsEnabled": False,
            "CloudWatchMetricsEnabled": False,
            "MetricName": "idk",
        }
    )
    wacl.should.have.key("Rules").length_of(2)


@mock_wafv2
def test_get_web_acl():
    conn = boto3.client("wafv2", region_name="us-east-1")
    body = CREATE_WEB_ACL_BODY("John", "REGIONAL")
    web_acl_id = conn.create_web_acl(**body)["Summary"]["Id"]
    wacl = conn.get_web_acl(Name="John", Scope="REGIONAL", Id=web_acl_id)["WebACL"]
    wacl.should.have.key("Name").equals("John")
    wacl.should.have.key("Id").equals(web_acl_id)


@mock_wafv2
def test_list_web_acl():
    conn = boto3.client("wafv2", region_name="us-east-1")
    conn.create_web_acl(**CREATE_WEB_ACL_BODY("Daphne", "REGIONAL"))
    conn.create_web_acl(**CREATE_WEB_ACL_BODY("Penelope", "CLOUDFRONT"))
    conn.create_web_acl(**CREATE_WEB_ACL_BODY("Sarah", "REGIONAL"))
    res = conn.list_web_acls(**LIST_WEB_ACL_BODY("REGIONAL"))
    web_acls = res["WebACLs"]
    assert len(web_acls) == 2
    assert web_acls[0]["Name"] == "Daphne"
    assert web_acls[1]["Name"] == "Sarah"

    res = conn.list_web_acls(**LIST_WEB_ACL_BODY("CLOUDFRONT"))
    web_acls = res["WebACLs"]
    assert len(web_acls) == 1
    assert web_acls[0]["Name"] == "Penelope"


@mock_wafv2
def test_delete_web_acl():
    conn = boto3.client("wafv2", region_name="us-east-1")
    wacl_id = conn.create_web_acl(**CREATE_WEB_ACL_BODY("Daphne", "REGIONAL"))[
        "Summary"
    ]["Id"]

    conn.delete_web_acl(Name="Daphne", Id=wacl_id, Scope="REGIONAL", LockToken="n/a")

    res = conn.list_web_acls(**LIST_WEB_ACL_BODY("REGIONAL"))
    res["WebACLs"].should.have.length_of(0)

    with pytest.raises(ClientError) as exc:
        conn.get_web_acl(Name="Daphne", Scope="REGIONAL", Id=wacl_id)
    err = exc.value.response["Error"]
    err["Code"].should.equal("WAFNonexistentItemException")
    err["Message"].should.equal(
        "AWS WAF couldn’t perform the operation because your resource doesn’t exist."
    )


@mock_wafv2
def test_update_web_acl():
    conn = boto3.client("wafv2", region_name="us-east-1")
    wacl_id = conn.create_web_acl(**CREATE_WEB_ACL_BODY("Daphne", "REGIONAL"))[
        "Summary"
    ]["Id"]

    resp = conn.update_web_acl(
        Name="Daphne",
        Scope="REGIONAL",
        Id=wacl_id,
        DefaultAction={"Block": {"CustomResponse": {"ResponseCode": 412}}},
        Description="updated_desc",
        Rules=[
            {
                "Name": "rule1",
                "Priority": 456,
                "Statement": {},
                "VisibilityConfig": {
                    "SampledRequestsEnabled": True,
                    "CloudWatchMetricsEnabled": True,
                    "MetricName": "updated",
                },
            }
        ],
        LockToken="n/a",
        VisibilityConfig={
            "SampledRequestsEnabled": True,
            "CloudWatchMetricsEnabled": True,
            "MetricName": "updated",
        },
    )
    resp.should.have.key("NextLockToken")

    acl = conn.get_web_acl(Name="Daphne", Scope="REGIONAL", Id=wacl_id)["WebACL"]
    acl.should.have.key("Description").equals("updated_desc")
    acl.should.have.key("DefaultAction").equals(
        {"Block": {"CustomResponse": {"ResponseCode": 412}}}
    )
    acl.should.have.key("Rules").equals(
        [
            {
                "Name": "rule1",
                "Priority": 456,
                "Statement": {},
                "VisibilityConfig": {
                    "SampledRequestsEnabled": True,
                    "CloudWatchMetricsEnabled": True,
                    "MetricName": "updated",
                },
            }
        ]
    )
    acl.should.have.key("VisibilityConfig").equals(
        {
            "SampledRequestsEnabled": True,
            "CloudWatchMetricsEnabled": True,
            "MetricName": "updated",
        }
    )



import json

from moto import settings
from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import ActionAuthenticatorMixin, BaseResponse
from typing import Any, Dict, List


class MotoAPIResponse(BaseResponse):
    def reset_response(
        self,
        request: Any,  # pylint: disable=unused-argument
        full_url: str,  # pylint: disable=unused-argument
        headers: Any,  # pylint: disable=unused-argument
    ) -> TYPE_RESPONSE:
        if request.method == "POST":
            from .models import moto_api_backend

            moto_api_backend.reset()
            return 200, {}, json.dumps({"status": "ok"})
        return 400, {}, json.dumps({"Error": "Need to POST to reset Moto"})

    def reset_auth_response(
        self,
        request: Any,  # pylint: disable=unused-argument
        full_url: str,  # pylint: disable=unused-argument
        headers: Any,  # pylint: disable=unused-argument
    ) -> TYPE_RESPONSE:
        if request.method == "POST":
            previous_initial_no_auth_action_count = (
                settings.INITIAL_NO_AUTH_ACTION_COUNT
            )
            settings.INITIAL_NO_AUTH_ACTION_COUNT = float(request.data.decode())
            ActionAuthenticatorMixin.request_count = 0
            return (
                200,
                {},
                json.dumps(
                    {
                        "status": "ok",
                        "PREVIOUS_INITIAL_NO_AUTH_ACTION_COUNT": str(
                            previous_initial_no_auth_action_count
                        ),
                    }
                ),
            )
        return 400, {}, json.dumps({"Error": "Need to POST to reset Moto Auth"})

    def model_data(
        self,
        request: Any,  # pylint: disable=unused-argument
        full_url: str,  # pylint: disable=unused-argument
        headers: Any,  # pylint: disable=unused-argument
    ) -> TYPE_RESPONSE:
        from moto.core.base_backend import model_data

        results: Dict[str, Dict[str, List[Any]]] = {}
        for service in sorted(model_data):
            models = model_data[service]
            results[service] = {}
            for name in sorted(models):
                model = models[name]
                results[service][name] = []
                for instance in model.instances:  # type: ignore[attr-defined]
                    inst_result = {}
                    for attr in dir(instance):
                        if not attr.startswith("_"):
                            try:
                                json.dumps(getattr(instance, attr))
                            except (TypeError, AttributeError):
                                pass
                            else:
                                inst_result[attr] = getattr(instance, attr)
                    results[service][name].append(inst_result)
        return 200, {"Content-Type": "application/javascript"}, json.dumps(results)

    def dashboard(
        self,
        request: Any,  # pylint: disable=unused-argument
        full_url: str,  # pylint: disable=unused-argument
        headers: Any,  # pylint: disable=unused-argument
    ) -> str:
        from flask import render_template

        return render_template("dashboard.html")

    def get_transition(
        self,
        request: Any,  # pylint: disable=unused-argument
        full_url: str,  # pylint: disable=unused-argument
        headers: Any,  # pylint: disable=unused-argument
    ) -> TYPE_RESPONSE:
        from .models import moto_api_backend

        qs_dict = dict(
            x.split("=") for x in request.query_string.decode("utf-8").split("&")
        )
        model_name = qs_dict["model_name"]

        resp = moto_api_backend.get_transition(model_name=model_name)

        return 200, {}, json.dumps(resp)

    def set_transition(
        self,
        request: Any,  # pylint: disable=unused-argument
        full_url: str,  # pylint: disable=unused-argument
        headers: Any,  # pylint: disable=unused-argument
    ) -> TYPE_RESPONSE:
        from .models import moto_api_backend

        request_body_size = int(headers["Content-Length"])
        body = request.environ["wsgi.input"].read(request_body_size).decode("utf-8")
        body = json.loads(body)
        model_name = body["model_name"]
        transition = body["transition"]

        moto_api_backend.set_transition(model_name, transition)
        return 201, {}, ""

    def unset_transition(
        self,
        request: Any,  # pylint: disable=unused-argument
        full_url: str,  # pylint: disable=unused-argument
        headers: Any,  # pylint: disable=unused-argument
    ) -> TYPE_RESPONSE:
        from .models import moto_api_backend

        request_body_size = int(headers["Content-Length"])
        body = request.environ["wsgi.input"].read(request_body_size).decode("utf-8")
        body = json.loads(body)
        model_name = body["model_name"]

        moto_api_backend.unset_transition(model_name)
        return 201, {}, ""

    def seed(self, req: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:
        self.setup_class(req, full_url, headers)
        from . import mock_random

        a = self._get_param("a")
        mock_random.seed(int(a))
        return 200, {}, ""



import boto3
import pytest
import sure  # noqa # pylint: disable=unused-import

from botocore.exceptions import ClientError
from moto import mock_apigateway

from .test_apigateway import create_method_integration


@mock_apigateway
def test_create_stage_minimal():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]

    create_method_integration(client, api_id)
    response = client.create_deployment(restApiId=api_id, stageName=stage_name)
    deployment_id = response["id"]

    new_stage_name = "current"
    response = client.create_stage(
        restApiId=api_id, stageName=new_stage_name, deploymentId=deployment_id
    )

    response.should.have.key("stageName").equals(new_stage_name)
    response.should.have.key("deploymentId").equals(deployment_id)
    response.should.have.key("methodSettings").equals({})
    response.should.have.key("variables").equals({})
    response.should.have.key("ResponseMetadata").should.have.key(
        "HTTPStatusCode"
    ).equals(201)
    response.should.have.key("description").equals("")
    response.shouldnt.have.key("cacheClusterStatus")
    response.should.have.key("cacheClusterEnabled").equals(False)

    stage = client.get_stage(restApiId=api_id, stageName=new_stage_name)
    stage["stageName"].should.equal(new_stage_name)
    stage["deploymentId"].should.equal(deployment_id)


@mock_apigateway
def test_create_stage_with_env_vars():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]

    create_method_integration(client, api_id)
    response = client.create_deployment(restApiId=api_id, stageName=stage_name)
    deployment_id = response["id"]

    new_stage_name_with_vars = "stage_with_vars"
    response = client.create_stage(
        restApiId=api_id,
        stageName=new_stage_name_with_vars,
        deploymentId=deployment_id,
        variables={"env": "dev"},
    )

    response.should.have.key("stageName").equals(new_stage_name_with_vars)
    response.should.have.key("deploymentId").equals(deployment_id)
    response.should.have.key("methodSettings").equals({})
    response.should.have.key("variables").equals({"env": "dev"})
    response.should.have.key("ResponseMetadata").should.have.key(
        "HTTPStatusCode"
    ).equals(201)
    response.should.have.key("description").equals("")
    response.shouldnt.have.key("cacheClusterStatus")
    response.should.have.key("cacheClusterEnabled").equals(False)

    stage = client.get_stage(restApiId=api_id, stageName=new_stage_name_with_vars)
    stage["stageName"].should.equal(new_stage_name_with_vars)
    stage["deploymentId"].should.equal(deployment_id)
    stage["variables"].should.have.key("env").which.should.match("dev")


@mock_apigateway
def test_create_stage_with_vars_and_cache():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]

    create_method_integration(client, api_id)
    response = client.create_deployment(restApiId=api_id, stageName=stage_name)
    deployment_id = response["id"]

    new_stage_name = "stage_with_vars_and_cache_settings"
    response = client.create_stage(
        restApiId=api_id,
        stageName=new_stage_name,
        deploymentId=deployment_id,
        variables={"env": "dev"},
        cacheClusterEnabled=True,
        description="hello moto",
    )

    response.should.have.key("stageName").equals(new_stage_name)
    response.should.have.key("deploymentId").equals(deployment_id)
    response.should.have.key("methodSettings").equals({})
    response.should.have.key("variables").equals({"env": "dev"})
    response.should.have.key("ResponseMetadata").should.have.key(
        "HTTPStatusCode"
    ).equals(201)
    response.should.have.key("description").equals("hello moto")
    response.should.have.key("cacheClusterStatus").equals("AVAILABLE")
    response.should.have.key("cacheClusterEnabled").equals(True)
    response.should.have.key("cacheClusterSize").equals("0.5")

    stage = client.get_stage(restApiId=api_id, stageName=new_stage_name)

    stage["cacheClusterSize"].should.equal("0.5")


@mock_apigateway
def test_create_stage_with_cache_settings():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]

    create_method_integration(client, api_id)
    response = client.create_deployment(restApiId=api_id, stageName=stage_name)
    deployment_id = response["id"]

    new_stage_name = "stage_with_vars_and_cache_settings_and_size"
    response = client.create_stage(
        restApiId=api_id,
        stageName=new_stage_name,
        deploymentId=deployment_id,
        variables={"env": "dev"},
        cacheClusterEnabled=True,
        cacheClusterSize="1.6",
        tracingEnabled=True,
        description="hello moto",
    )

    response.should.have.key("stageName").equals(new_stage_name)
    response.should.have.key("deploymentId").equals(deployment_id)
    response.should.have.key("methodSettings").equals({})
    response.should.have.key("variables").equals({"env": "dev"})
    response.should.have.key("ResponseMetadata").should.have.key(
        "HTTPStatusCode"
    ).equals(201)
    response.should.have.key("description").equals("hello moto")
    response.should.have.key("cacheClusterStatus").equals("AVAILABLE")
    response.should.have.key("cacheClusterEnabled").equals(True)
    response.should.have.key("cacheClusterSize").equals("1.6")
    response.should.have.key("tracingEnabled").equals(True)

    stage = client.get_stage(restApiId=api_id, stageName=new_stage_name)
    stage["stageName"].should.equal(new_stage_name)
    stage["deploymentId"].should.equal(deployment_id)
    stage["variables"].should.have.key("env").which.should.match("dev")
    stage["cacheClusterSize"].should.equal("1.6")


@mock_apigateway
def test_recreate_stage_from_deployment():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]
    create_method_integration(client, api_id)
    depl_id1 = client.create_deployment(restApiId=api_id, stageName=stage_name)["id"]

    with pytest.raises(ClientError) as exc:
        client.create_stage(
            restApiId=api_id, stageName=stage_name, deploymentId=depl_id1
        )
    err = exc.value.response["Error"]
    err["Code"].should.equal("ConflictException")
    err["Message"].should.equal("Stage already exists")


@mock_apigateway
def test_create_stage_twice():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]
    create_method_integration(client, api_id)
    depl_id1 = client.create_deployment(restApiId=api_id, stageName=stage_name)["id"]

    new_stage_name = "current"
    client.create_stage(
        restApiId=api_id, stageName=new_stage_name, deploymentId=depl_id1
    )

    with pytest.raises(ClientError) as exc:
        client.create_stage(
            restApiId=api_id, stageName=new_stage_name, deploymentId=depl_id1
        )
    err = exc.value.response["Error"]
    err["Code"].should.equal("ConflictException")
    err["Message"].should.equal("Stage already exists")


@mock_apigateway
def test_delete_stage():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]
    create_method_integration(client, api_id)
    depl_id1 = client.create_deployment(restApiId=api_id, stageName=stage_name)["id"]
    depl_id2 = client.create_deployment(restApiId=api_id, stageName=stage_name)["id"]

    new_stage_name = "current"
    client.create_stage(
        restApiId=api_id, stageName=new_stage_name, deploymentId=depl_id1
    )

    new_stage_name_with_vars = "stage_with_vars"
    client.create_stage(
        restApiId=api_id,
        stageName=new_stage_name_with_vars,
        deploymentId=depl_id2,
        variables={"env": "dev"},
    )
    stages = client.get_stages(restApiId=api_id)["item"]
    stage_names = [stage["stageName"] for stage in stages]
    stage_names.should.have.length_of(3)
    stage_names.should.contain(stage_name)
    stage_names.should.contain(new_stage_name)
    stage_names.should.contain(new_stage_name_with_vars)

    # delete stage
    response = client.delete_stage(restApiId=api_id, stageName=new_stage_name_with_vars)
    response["ResponseMetadata"]["HTTPStatusCode"].should.equal(202)

    # verify other stage still exists
    stages = client.get_stages(restApiId=api_id)["item"]
    stage_names = [stage["stageName"] for stage in stages]
    stage_names.should.have.length_of(2)
    stage_names.shouldnt.contain(new_stage_name_with_vars)


@mock_apigateway
def test_delete_stage_created_by_deployment():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]
    create_method_integration(client, api_id)
    depl_id1 = client.create_deployment(restApiId=api_id, stageName=stage_name)["id"]

    # Sanity check that the deployment exists
    depls = client.get_deployments(restApiId=api_id)["items"]
    depls.should.have.length_of(1)
    set(depls[0].keys()).should.equal({"id", "createdDate"})

    # Sanity check that the stage exists
    stage = client.get_stages(restApiId=api_id)["item"][0]
    stage.should.have.key("deploymentId").equals(depl_id1)
    stage.should.have.key("stageName").equals(stage_name)

    # delete stage
    response = client.delete_stage(restApiId=api_id, stageName=stage_name)
    response["ResponseMetadata"]["HTTPStatusCode"].should.equal(202)

    # verify no stage exists
    stages = client.get_stages(restApiId=api_id)
    stages.should.have.key("item").equals([])

    # verify deployment still exists, unchanged
    depls = client.get_deployments(restApiId=api_id)["items"]
    depls.should.have.length_of(1)
    set(depls[0].keys()).should.equal({"id", "createdDate"})


@mock_apigateway
def test_delete_stage_unknown_stage():
    client = boto3.client("apigateway", region_name="us-west-2")
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]
    with pytest.raises(ClientError) as exc:
        client.delete_stage(restApiId=api_id, stageName="unknown")
    err = exc.value.response["Error"]
    err["Message"].should.equal("Invalid stage identifier specified")
    err["Code"].should.equal("NotFoundException")


@mock_apigateway
def test_update_stage_configuration():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]
    create_method_integration(client, api_id)

    response = client.create_deployment(
        restApiId=api_id, stageName=stage_name, description="1.0.1"
    )
    deployment_id = response["id"]

    response = client.get_deployment(restApiId=api_id, deploymentId=deployment_id)

    response.should.have.key("id").equals(deployment_id)
    response.should.have.key("ResponseMetadata").should.have.key(
        "HTTPStatusCode"
    ).equals(200)
    response.should.have.key("description").equals("1.0.1")

    response = client.create_deployment(
        restApiId=api_id, stageName=stage_name, description="1.0.2"
    )
    deployment_id2 = response["id"]

    stage = client.get_stage(restApiId=api_id, stageName=stage_name)
    stage["stageName"].should.equal(stage_name)
    stage["deploymentId"].should.equal(deployment_id2)
    stage.shouldnt.have.key("cacheClusterSize")
    stage.shouldnt.have.key("cacheClusterStatus")

    client.update_stage(
        restApiId=api_id,
        stageName=stage_name,
        patchOperations=[
            {"op": "replace", "path": "/cacheClusterEnabled", "value": "True"}
        ],
    )

    stage = client.get_stage(restApiId=api_id, stageName=stage_name)

    stage.should.have.key("cacheClusterSize").which.should.equal("0.5")
    stage.should.have.key("cacheClusterStatus").equals("AVAILABLE")

    client.update_stage(
        restApiId=api_id,
        stageName=stage_name,
        patchOperations=[
            {"op": "replace", "path": "/cacheClusterSize", "value": "1.6"}
        ],
    )

    stage = client.get_stage(restApiId=api_id, stageName=stage_name)

    stage.should.have.key("cacheClusterSize").which.should.equal("1.6")

    client.update_stage(
        restApiId=api_id,
        stageName=stage_name,
        patchOperations=[
            {"op": "replace", "path": "/deploymentId", "value": deployment_id},
            {"op": "replace", "path": "/variables/environment", "value": "dev"},
            {"op": "replace", "path": "/variables/region", "value": "eu-west-1"},
            {"op": "replace", "path": "/*/*/caching/dataEncrypted", "value": "True"},
            {"op": "replace", "path": "/cacheClusterEnabled", "value": "True"},
            {
                "op": "replace",
                "path": "/description",
                "value": "stage description update",
            },
            {"op": "replace", "path": "/cacheClusterSize", "value": "1.6"},
        ],
    )

    client.update_stage(
        restApiId=api_id,
        stageName=stage_name,
        patchOperations=[
            {"op": "remove", "path": "/variables/region", "value": "eu-west-1"}
        ],
    )

    stage = client.get_stage(restApiId=api_id, stageName=stage_name)

    stage["description"].should.match("stage description update")
    stage["cacheClusterSize"].should.equal("1.6")
    stage["variables"]["environment"].should.match("dev")
    stage["variables"].should_not.have.key("region")
    stage["cacheClusterEnabled"].should.equal(True)
    stage["deploymentId"].should.match(deployment_id)
    stage["methodSettings"].should.have.key("*/*")
    stage["methodSettings"]["*/*"].should.have.key(
        "cacheDataEncrypted"
    ).which.should.be.true


@mock_apigateway
def test_update_stage_add_access_log_settings():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]
    create_method_integration(client, api_id)

    client.create_deployment(
        restApiId=api_id, stageName=stage_name, description="1.0.1"
    )

    client.update_stage(
        restApiId=api_id,
        stageName=stage_name,
        patchOperations=[
            {
                "op": "replace",
                "path": "/accessLogSettings/destinationArn",
                "value": "arn:aws:logs:us-east-1:123456789012:log-group:foo-bar-x0hyv",
            },
            {
                "op": "replace",
                "path": "/accessLogSettings/format",
                "value": "$context.identity.sourceIp msg",
            },
        ],
    )

    stage = client.get_stage(restApiId=api_id, stageName=stage_name)
    stage.should.have.key("accessLogSettings").equals(
        {
            "format": "$context.identity.sourceIp msg",
            "destinationArn": "arn:aws:logs:us-east-1:123456789012:log-group:foo-bar-x0hyv",
        }
    )


@mock_apigateway
def test_update_stage_tracing_disabled():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]
    create_method_integration(client, api_id)
    response = client.create_deployment(restApiId=api_id, stageName=stage_name)

    client.update_stage(
        restApiId=api_id,
        stageName=stage_name,
        patchOperations=[
            {"op": "replace", "path": "/tracingEnabled", "value": "false"}
        ],
    )

    stage = client.get_stage(restApiId=api_id, stageName=stage_name)
    stage.should.have.key("tracingEnabled").equals(False)

    client.update_stage(
        restApiId=api_id,
        stageName=stage_name,
        patchOperations=[{"op": "replace", "path": "/tracingEnabled", "value": "true"}],
    )

    stage = client.get_stage(restApiId=api_id, stageName=stage_name)
    stage.should.have.key("tracingEnabled").equals(True)


@mock_apigateway
def test_update_stage_remove_access_log_settings():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]
    create_method_integration(client, api_id)

    client.create_deployment(
        restApiId=api_id, stageName=stage_name, description="1.0.1"
    )

    client.update_stage(
        restApiId=api_id,
        stageName=stage_name,
        patchOperations=[{"op": "remove", "path": "/accessLogSettings"}],
    )

    stage = client.get_stage(restApiId=api_id, stageName=stage_name)
    stage.shouldnt.have.key("accessLogSettings")


@mock_apigateway
def test_update_stage_configuration_unknown_operation():
    client = boto3.client("apigateway", region_name="us-west-2")
    stage_name = "staging"
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]
    create_method_integration(client, api_id)

    client.create_deployment(
        restApiId=api_id, stageName=stage_name, description="1.0.1"
    )

    with pytest.raises(ClientError) as exc:
        client.update_stage(
            restApiId=api_id,
            stageName=stage_name,
            patchOperations=[
                {"op": "unknown_op", "path": "/notasetting", "value": "eu-west-1"}
            ],
        )
    err = exc.value.response["Error"]
    err["Code"].should.equal("ValidationException")
    err["Message"].should.equal(
        "Member must satisfy enum value set: [add, remove, move, test, replace, copy]"
    )


@mock_apigateway
def test_non_existent_stage():
    client = boto3.client("apigateway", region_name="us-west-2")
    response = client.create_rest_api(name="my_api", description="this is my api")
    api_id = response["id"]

    with pytest.raises(ClientError) as exc:
        client.get_stage(restApiId=api_id, stageName="xxx")
    err = exc.value.response["Error"]
    err["Code"].should.equal("NotFoundException")



from moto.core.exceptions import JsonRESTError


class DataBrewClientError(JsonRESTError):
    code = 400


class AlreadyExistsException(DataBrewClientError):
    def __init__(self, typ):
        super().__init__("AlreadyExistsException", f"{typ} already exists.")


class ConflictException(DataBrewClientError):
    code = 409

    def __init__(self, message, **kwargs):
        super().__init__("ConflictException", message, **kwargs)


class ValidationException(DataBrewClientError):
    def __init__(self, message, **kwargs):
        super().__init__("ValidationException", message, **kwargs)


class RulesetAlreadyExistsException(AlreadyExistsException):
    def __init__(self):
        super().__init__("Ruleset")


class EntityNotFoundException(DataBrewClientError):
    def __init__(self, msg):
        super().__init__("EntityNotFoundException", msg)


class ResourceNotFoundException(DataBrewClientError):
    code = 404

    def __init__(self, message, **kwargs):
        super().__init__("ResourceNotFoundException", message, **kwargs)


class RulesetNotFoundException(EntityNotFoundException):
    def __init__(self, recipe_name):
        super().__init__(f"Ruleset {recipe_name} not found.")


class ServiceQuotaExceededException(JsonRESTError):
    code = 402

    def __init__(self):
        super().__init__(
            "ServiceQuotaExceededException", "A service quota is exceeded."
        )



import boto3
import pytest
import sure  # noqa # pylint: disable=unused-import

from botocore.exceptions import ClientError
from moto import mock_s3
from uuid import uuid4

DEFAULT_REGION_NAME = "us-east-1"


@mock_s3
def test_get_bucket_replication_for_unexisting_bucket():
    bucket_name = str(uuid4())
    s3 = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    with pytest.raises(ClientError) as exc:
        s3.get_bucket_replication(Bucket=bucket_name)
    err = exc.value.response["Error"]
    err["Code"].should.equal("NoSuchBucket")
    err["Message"].should.equal("The specified bucket does not exist")
    err["BucketName"].should.equal(bucket_name)


@mock_s3
def test_get_bucket_replication_bucket_without_replication():
    bucket_name = str(uuid4())
    s3 = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3.create_bucket(Bucket=bucket_name)

    with pytest.raises(ClientError) as exc:
        s3.get_bucket_replication(Bucket=bucket_name)
    err = exc.value.response["Error"]
    err["Code"].should.equal("ReplicationConfigurationNotFoundError")
    err["Message"].should.equal("The replication configuration was not found")
    err["BucketName"].should.equal(bucket_name)


@mock_s3
def test_delete_bucket_replication_unknown_bucket():
    bucket_name = str(uuid4())
    s3 = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    with pytest.raises(ClientError) as exc:
        s3.delete_bucket_replication(Bucket=bucket_name)
    err = exc.value.response["Error"]
    err["Code"].should.equal("NoSuchBucket")
    err["Message"].should.equal("The specified bucket does not exist")
    err["BucketName"].should.equal(bucket_name)


@mock_s3
def test_delete_bucket_replication_bucket_without_replication():
    bucket_name = str(uuid4())
    s3 = boto3.client("s3", region_name=DEFAULT_REGION_NAME)

    s3.create_bucket(Bucket=bucket_name)
    # No-op
    s3.delete_bucket_replication(Bucket=bucket_name)


@mock_s3
def test_create_replication_without_versioning():
    bucket_name = str(uuid4())
    s3 = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3.create_bucket(Bucket=bucket_name)

    with pytest.raises(ClientError) as exc:
        s3.put_bucket_replication(
            Bucket=bucket_name,
            ReplicationConfiguration={
                "Role": "myrole",
                "Rules": [
                    {"Destination": {"Bucket": "secondbucket"}, "Status": "Enabled"}
                ],
            },
        )
    err = exc.value.response["Error"]
    err["Code"].should.equal("InvalidRequest")
    err["Message"].should.equal(
        "Versioning must be 'Enabled' on the bucket to apply a replication configuration"
    )
    err["BucketName"].should.equal(bucket_name)


@mock_s3
def test_create_and_retrieve_replication_with_single_rules():
    bucket_name = str(uuid4())
    s3 = boto3.client("s3", region_name=DEFAULT_REGION_NAME)

    s3.create_bucket(Bucket=bucket_name)
    s3.put_bucket_versioning(
        Bucket=bucket_name, VersioningConfiguration={"Status": "Enabled"}
    )
    s3.put_bucket_replication(
        Bucket=bucket_name,
        ReplicationConfiguration={
            "Role": "myrole",
            "Rules": [
                {
                    "ID": "firstrule",
                    "Priority": 2,
                    "Destination": {"Bucket": "secondbucket"},
                    "Status": "Enabled",
                }
            ],
        },
    )

    config = s3.get_bucket_replication(Bucket=bucket_name)["ReplicationConfiguration"]
    config.should.equal(
        {
            "Role": "myrole",
            "Rules": [
                {
                    "DeleteMarkerReplication": {"Status": "Disabled"},
                    "Destination": {"Bucket": "secondbucket"},
                    "Filter": {"Prefix": ""},
                    "ID": "firstrule",
                    "Priority": 2,
                    "Status": "Enabled",
                }
            ],
        }
    )

    s3.delete_bucket_replication(Bucket=bucket_name)

    # Can't retrieve replication that has been deleted
    with pytest.raises(ClientError) as exc:
        s3.get_bucket_replication(Bucket=bucket_name)
    err = exc.value.response["Error"]
    err["Code"].should.equal("ReplicationConfigurationNotFoundError")
    err["Message"].should.equal("The replication configuration was not found")
    err["BucketName"].should.equal(bucket_name)


@mock_s3
def test_create_and_retrieve_replication_with_multiple_rules():
    bucket_name = str(uuid4())
    s3 = boto3.client("s3", region_name=DEFAULT_REGION_NAME)

    s3.create_bucket(Bucket=bucket_name)
    s3.put_bucket_versioning(
        Bucket=bucket_name, VersioningConfiguration={"Status": "Enabled"}
    )
    s3.put_bucket_replication(
        Bucket=bucket_name,
        ReplicationConfiguration={
            "Role": "myrole",
            "Rules": [
                {"Destination": {"Bucket": "secondbucket"}, "Status": "Enabled"},
                {
                    "ID": "secondrule",
                    "Priority": 2,
                    "Destination": {"Bucket": "thirdbucket"},
                    "Status": "Disabled",
                },
            ],
        },
    )

    config = s3.get_bucket_replication(Bucket=bucket_name)["ReplicationConfiguration"]
    config.should.have.key("Role").equal("myrole")
    rules = config["Rules"]
    rules.should.have.length_of(2)

    first_rule = rules[0]
    first_rule.should.have.key("ID")
    first_rule.should.have.key("Priority").equal(1)
    first_rule.should.have.key("Status").equal("Enabled")
    first_rule.should.have.key("Destination").equal({"Bucket": "secondbucket"})

    second = rules[1]
    second.should.have.key("ID").equal("secondrule")
    second.should.have.key("Priority").equal(2)
    second.should.have.key("Status").equal("Disabled")
    second.should.have.key("Destination").equal({"Bucket": "thirdbucket"})



from moto.core import BaseBackend, BackendDict, BaseModel
from moto.ec2 import ec2_backends
from moto.moto_api._internal import mock_random as random
import datetime

from .exceptions import ResourceNotFoundException, ValidationException


class OpsworkInstance(BaseModel):
    """
    opsworks maintains its own set of ec2 instance metadata.
    This metadata exists before any instance reservations are made, and is
    used to populate a reservation request when "start" is called
    """

    def __init__(
        self,
        stack_id,
        layer_ids,
        instance_type,
        ec2_backend,
        auto_scale_type=None,
        hostname=None,
        os=None,
        ami_id="ami-08111162",
        ssh_keyname=None,
        availability_zone=None,
        virtualization_type="hvm",
        subnet_id=None,
        architecture="x86_64",
        root_device_type="ebs",
        block_device_mappings=None,
        install_updates_on_boot=True,
        ebs_optimized=False,
        agent_version="INHERIT",
        instance_profile_arn=None,
        associate_public_ip=None,
        security_group_ids=None,
    ):

        self.ec2_backend = ec2_backend

        self.instance_profile_arn = instance_profile_arn
        self.agent_version = agent_version
        self.ebs_optimized = ebs_optimized
        self.install_updates_on_boot = install_updates_on_boot
        self.architecture = architecture
        self.virtualization_type = virtualization_type
        self.ami_id = ami_id
        self.auto_scale_type = auto_scale_type
        self.instance_type = instance_type
        self.layer_ids = layer_ids
        self.stack_id = stack_id

        # may not be totally accurate defaults; instance-type dependent
        self.root_device_type = root_device_type
        # todo: refactor how we track block_device_mappings to use
        # boto.ec2.blockdevicemapping.BlockDeviceType and standardize
        # formatting in to_dict()
        self.block_device_mappings = block_device_mappings
        if self.block_device_mappings is None:
            self.block_device_mappings = [
                {
                    "DeviceName": "ROOT_DEVICE",
                    "Ebs": {"VolumeSize": 8, "VolumeType": "gp2"},
                }
            ]
        self.security_group_ids = security_group_ids
        if self.security_group_ids is None:
            self.security_group_ids = []

        self.os = os
        self.hostname = hostname
        self.ssh_keyname = ssh_keyname
        self.availability_zone = availability_zone
        self.subnet_id = subnet_id
        self.associate_public_ip = associate_public_ip

        self.instance = None
        self.reported_os = {}
        self.infrastructure_class = "ec2 (fixed)"
        self.platform = "linux (fixed)"

        self.id = "{0}".format(random.uuid4())
        self.created_at = datetime.datetime.utcnow()

    def start(self):
        """
        create an ec2 reservation if one doesn't already exist and call
        start_instance. Update instance attributes to the newly created instance
        attributes
        """
        if self.instance is None:
            reservation = self.ec2_backend.add_instances(
                image_id=self.ami_id,
                count=1,
                user_data="",
                security_group_names=[],
                security_group_ids=self.security_group_ids,
                instance_type=self.instance_type,
                is_instance_type_default=not self.instance_type,
                key_name=self.ssh_keyname,
                ebs_optimized=self.ebs_optimized,
                subnet_id=self.subnet_id,
                associate_public_ip=self.associate_public_ip,
            )
            self.instance = reservation.instances[0]
            self.reported_os = {
                "Family": "rhel (fixed)",
                "Name": "amazon (fixed)",
                "Version": "2016.03 (fixed)",
            }
            self.platform = self.instance.platform
            self.security_group_ids = self.instance.security_groups
            self.architecture = self.instance.architecture
            self.virtualization_type = self.instance.virtualization_type
            self.subnet_id = self.instance.subnet_id
            self.root_device_type = self.instance.root_device_type

        self.ec2_backend.start_instances([self.instance.id])

    @property
    def status(self):
        if self.instance is None:
            return "stopped"
        # OpsWorks reports the "running" state as "online"
        elif self.instance._state.name == "running":
            return "online"
        return self.instance._state.name

    def to_dict(self):
        d = {
            "AgentVersion": self.agent_version,
            "Architecture": self.architecture,
            "AvailabilityZone": self.availability_zone,
            "BlockDeviceMappings": self.block_device_mappings,
            "CreatedAt": self.created_at.isoformat(),
            "EbsOptimized": self.ebs_optimized,
            "InstanceId": self.id,
            "Hostname": self.hostname,
            "InfrastructureClass": self.infrastructure_class,
            "InstallUpdatesOnBoot": self.install_updates_on_boot,
            "InstanceProfileArn": self.instance_profile_arn,
            "InstanceType": self.instance_type,
            "LayerIds": self.layer_ids,
            "Os": self.os,
            "Platform": self.platform,
            "ReportedOs": self.reported_os,
            "RootDeviceType": self.root_device_type,
            "SecurityGroupIds": self.security_group_ids,
            "AmiId": self.ami_id,
            "Status": self.status,
        }
        if self.ssh_keyname is not None:
            d.update({"SshKeyName": self.ssh_keyname})

        if self.auto_scale_type is not None:
            d.update({"AutoScaleType": self.auto_scale_type})

        if self.instance is not None:
            d.update({"Ec2InstanceId": self.instance.id})
            d.update({"ReportedAgentVersion": "2425-20160406102508 (fixed)"})
            d.update({"RootDeviceVolumeId": "vol-a20e450a (fixed)"})
            if self.ssh_keyname is not None:
                d.update(
                    {
                        "SshHostDsaKeyFingerprint": "24:36:32:fe:d8:5f:9c:18:b1:ad:37:e9:eb:e8:69:58 (fixed)"
                    }
                )
                d.update(
                    {
                        "SshHostRsaKeyFingerprint": "3c:bd:37:52:d7:ca:67:e1:6e:4b:ac:31:86:79:f5:6c (fixed)"
                    }
                )
            d.update({"PrivateDns": self.instance.private_dns})
            d.update({"PrivateIp": self.instance.private_ip})
            d.update({"PublicDns": getattr(self.instance, "public_dns", None)})
            d.update({"PublicIp": getattr(self.instance, "public_ip", None)})
        return d


class Layer(BaseModel):
    def __init__(
        self,
        stack_id,
        layer_type,
        name,
        shortname,
        attributes=None,
        custom_instance_profile_arn=None,
        custom_json=None,
        custom_security_group_ids=None,
        packages=None,
        volume_configurations=None,
        enable_autohealing=None,
        auto_assign_elastic_ips=None,
        auto_assign_public_ips=None,
        custom_recipes=None,
        install_updates_on_boot=None,
        use_ebs_optimized_instances=None,
        lifecycle_event_configuration=None,
    ):
        self.stack_id = stack_id
        self.type = layer_type
        self.name = name
        self.shortname = shortname

        self.attributes = attributes
        if attributes is None:
            self.attributes = {
                "BundlerVersion": None,
                "EcsClusterArn": None,
                "EnableHaproxyStats": None,
                "GangliaPassword": None,
                "GangliaUrl": None,
                "GangliaUser": None,
                "HaproxyHealthCheckMethod": None,
                "HaproxyHealthCheckUrl": None,
                "HaproxyStatsPassword": None,
                "HaproxyStatsUrl": None,
                "HaproxyStatsUser": None,
                "JavaAppServer": None,
                "JavaAppServerVersion": None,
                "Jvm": None,
                "JvmOptions": None,
                "JvmVersion": None,
                "ManageBundler": None,
                "MemcachedMemory": None,
                "MysqlRootPassword": None,
                "MysqlRootPasswordUbiquitous": None,
                "NodejsVersion": None,
                "PassengerVersion": None,
                "RailsStack": None,
                "RubyVersion": None,
                "RubygemsVersion": None,
            }  # May not be accurate

        self.packages = packages
        if packages is None:
            self.packages = packages

        self.custom_recipes = custom_recipes
        if custom_recipes is None:
            self.custom_recipes = {
                "Configure": [],
                "Deploy": [],
                "Setup": [],
                "Shutdown": [],
                "Undeploy": [],
            }

        self.custom_security_group_ids = custom_security_group_ids
        if custom_security_group_ids is None:
            self.custom_security_group_ids = []

        self.lifecycle_event_configuration = lifecycle_event_configuration
        if lifecycle_event_configuration is None:
            self.lifecycle_event_configuration = {
                "Shutdown": {"DelayUntilElbConnectionsDrained": False}
            }

        self.volume_configurations = volume_configurations
        if volume_configurations is None:
            self.volume_configurations = []

        self.custom_instance_profile_arn = custom_instance_profile_arn
        self.custom_json = custom_json
        self.enable_autohealing = enable_autohealing
        self.auto_assign_elastic_ips = auto_assign_elastic_ips
        self.auto_assign_public_ips = auto_assign_public_ips
        self.install_updates_on_boot = install_updates_on_boot
        self.use_ebs_optimized_instances = use_ebs_optimized_instances

        self.id = "{0}".format(random.uuid4())
        self.created_at = datetime.datetime.utcnow()

    def __eq__(self, other):
        return self.id == other.id

    def to_dict(self):
        d = {
            "Attributes": self.attributes,
            "AutoAssignElasticIps": self.auto_assign_elastic_ips,
            "AutoAssignPublicIps": self.auto_assign_public_ips,
            "CreatedAt": self.created_at.isoformat(),
            "CustomRecipes": self.custom_recipes,
            "CustomSecurityGroupIds": self.custom_security_group_ids,
            "DefaultRecipes": {
                "Configure": [],
                "Setup": [],
                "Shutdown": [],
                "Undeploy": [],
            },  # May not be accurate
            "DefaultSecurityGroupNames": ["AWS-OpsWorks-Custom-Server"],
            "EnableAutoHealing": self.enable_autohealing,
            "LayerId": self.id,
            "LifecycleEventConfiguration": self.lifecycle_event_configuration,
            "Name": self.name,
            "Shortname": self.shortname,
            "StackId": self.stack_id,
            "Type": self.type,
            "UseEbsOptimizedInstances": self.use_ebs_optimized_instances,
            "VolumeConfigurations": self.volume_configurations,
        }
        if self.custom_json is not None:
            d.update({"CustomJson": self.custom_json})
        if self.custom_instance_profile_arn is not None:
            d.update({"CustomInstanceProfileArn": self.custom_instance_profile_arn})
        return d


class Stack(BaseModel):
    def __init__(
        self,
        name,
        account_id,
        region,
        service_role_arn,
        default_instance_profile_arn,
        vpcid="vpc-1f99bf7a",
        attributes=None,
        default_os="Ubuntu 12.04 LTS",
        hostname_theme="Layer_Dependent",
        default_availability_zone="us-east-1a",
        default_subnet_id="subnet-73981004",
        custom_json=None,
        configuration_manager=None,
        chef_configuration=None,
        use_custom_cookbooks=False,
        use_opsworks_security_groups=True,
        custom_cookbooks_source=None,
        default_ssh_keyname=None,
        default_root_device_type="instance-store",
        agent_version="LATEST",
    ):

        self.name = name
        self.region = region
        self.service_role_arn = service_role_arn
        self.default_instance_profile_arn = default_instance_profile_arn

        self.vpcid = vpcid
        self.attributes = attributes
        if attributes is None:
            self.attributes = {"Color": None}

        self.configuration_manager = configuration_manager
        if configuration_manager is None:
            self.configuration_manager = {"Name": "Chef", "Version": "11.4"}

        self.chef_configuration = chef_configuration
        if chef_configuration is None:
            self.chef_configuration = {}

        self.custom_cookbooks_source = custom_cookbooks_source
        if custom_cookbooks_source is None:
            self.custom_cookbooks_source = {}

        self.custom_json = custom_json
        self.default_ssh_keyname = default_ssh_keyname
        self.default_os = default_os
        self.hostname_theme = hostname_theme
        self.default_availability_zone = default_availability_zone
        self.default_subnet_id = default_subnet_id
        self.use_custom_cookbooks = use_custom_cookbooks
        self.use_opsworks_security_groups = use_opsworks_security_groups
        self.default_root_device_type = default_root_device_type
        self.agent_version = agent_version

        self.id = "{0}".format(random.uuid4())
        self.layers = []
        self.apps = []
        self.account_number = account_id
        self.created_at = datetime.datetime.utcnow()

    def __eq__(self, other):
        return self.id == other.id

    def generate_hostname(self):
        # this doesn't match amazon's implementation
        return "{theme}-{rand}-(moto)".format(
            theme=self.hostname_theme,
            rand=[random.choice("abcdefghijhk") for _ in range(4)],
        )

    @property
    def arn(self):
        return "arn:aws:opsworks:{region}:{account_number}:stack/{id}".format(
            region=self.region, account_number=self.account_number, id=self.id
        )

    def to_dict(self):
        response = {
            "AgentVersion": self.agent_version,
            "Arn": self.arn,
            "Attributes": self.attributes,
            "ChefConfiguration": self.chef_configuration,
            "ConfigurationManager": self.configuration_manager,
            "CreatedAt": self.created_at.isoformat(),
            "CustomCookbooksSource": self.custom_cookbooks_source,
            "DefaultAvailabilityZone": self.default_availability_zone,
            "DefaultInstanceProfileArn": self.default_instance_profile_arn,
            "DefaultOs": self.default_os,
            "DefaultRootDeviceType": self.default_root_device_type,
            "DefaultSshKeyName": self.default_ssh_keyname,
            "DefaultSubnetId": self.default_subnet_id,
            "HostnameTheme": self.hostname_theme,
            "Name": self.name,
            "Region": self.region,
            "ServiceRoleArn": self.service_role_arn,
            "StackId": self.id,
            "UseCustomCookbooks": self.use_custom_cookbooks,
            "UseOpsworksSecurityGroups": self.use_opsworks_security_groups,
            "VpcId": self.vpcid,
        }
        if self.custom_json is not None:
            response.update({"CustomJson": self.custom_json})
        if self.default_ssh_keyname is not None:
            response.update({"DefaultSshKeyName": self.default_ssh_keyname})
        return response


class App(BaseModel):
    def __init__(
        self,
        stack_id,
        name,
        app_type,
        shortname=None,
        description=None,
        datasources=None,
        app_source=None,
        domains=None,
        enable_ssl=False,
        ssl_configuration=None,
        attributes=None,
        environment=None,
    ):
        self.stack_id = stack_id
        self.name = name
        self.type = app_type
        self.shortname = shortname
        self.description = description

        self.datasources = datasources
        if datasources is None:
            self.datasources = []

        self.app_source = app_source
        if app_source is None:
            self.app_source = {}

        self.domains = domains
        if domains is None:
            self.domains = []

        self.enable_ssl = enable_ssl

        self.ssl_configuration = ssl_configuration
        if ssl_configuration is None:
            self.ssl_configuration = {}

        self.attributes = attributes
        if attributes is None:
            self.attributes = {}

        self.environment = environment
        if environment is None:
            self.environment = {}

        self.id = "{0}".format(random.uuid4())
        self.created_at = datetime.datetime.utcnow()

    def __eq__(self, other):
        return self.id == other.id

    def to_dict(self):
        d = {
            "AppId": self.id,
            "AppSource": self.app_source,
            "Attributes": self.attributes,
            "CreatedAt": self.created_at.isoformat(),
            "Datasources": self.datasources,
            "Description": self.description,
            "Domains": self.domains,
            "EnableSsl": self.enable_ssl,
            "Environment": self.environment,
            "Name": self.name,
            "Shortname": self.shortname,
            "SslConfiguration": self.ssl_configuration,
            "StackId": self.stack_id,
            "Type": self.type,
        }
        return d


class OpsWorksBackend(BaseBackend):
    def __init__(self, region_name, account_id):
        super().__init__(region_name, account_id)
        self.stacks = {}
        self.layers = {}
        self.apps = {}
        self.instances = {}
        self.ec2_backend = ec2_backends[account_id][region_name]

    def create_stack(self, **kwargs):
        stack = Stack(account_id=self.account_id, **kwargs)
        self.stacks[stack.id] = stack
        return stack

    def create_layer(self, **kwargs):
        name = kwargs["name"]
        shortname = kwargs["shortname"]
        stackid = kwargs["stack_id"]
        if stackid not in self.stacks:
            raise ResourceNotFoundException(stackid)
        if name in [layer.name for layer in self.stacks[stackid].layers]:
            raise ValidationException(
                'There is already a layer named "{0}" ' "for this stack".format(name)
            )
        if shortname in [layer.shortname for layer in self.stacks[stackid].layers]:
            raise ValidationException(
                'There is already a layer with shortname "{0}" '
                "for this stack".format(shortname)
            )
        layer = Layer(**kwargs)
        self.layers[layer.id] = layer
        self.stacks[stackid].layers.append(layer)
        return layer

    def create_app(self, **kwargs):
        name = kwargs["name"]
        stackid = kwargs["stack_id"]
        if stackid not in self.stacks:
            raise ResourceNotFoundException(stackid)
        if name in [a.name for a in self.stacks[stackid].apps]:
            raise ValidationException(
                'There is already an app named "{0}" ' "for this stack".format(name)
            )
        app = App(**kwargs)
        self.apps[app.id] = app
        self.stacks[stackid].apps.append(app)
        return app

    def create_instance(self, **kwargs):
        stack_id = kwargs["stack_id"]
        layer_ids = kwargs["layer_ids"]

        if stack_id not in self.stacks:
            raise ResourceNotFoundException(
                "Unable to find stack with ID {0}".format(stack_id)
            )

        unknown_layers = set(layer_ids) - set(self.layers.keys())
        if unknown_layers:
            raise ResourceNotFoundException(", ".join(unknown_layers))

        layers = [self.layers[id] for id in layer_ids]
        if len(set([layer.stack_id for layer in layers])) != 1 or any(
            [layer.stack_id != stack_id for layer in layers]
        ):
            raise ValidationException(
                "Please only provide layer IDs from the same stack"
            )

        stack = self.stacks[stack_id]
        # pick the first to set default instance_profile_arn and
        # security_group_ids on the instance.
        layer = layers[0]

        kwargs.setdefault("hostname", stack.generate_hostname())
        kwargs.setdefault("ssh_keyname", stack.default_ssh_keyname)
        kwargs.setdefault("availability_zone", stack.default_availability_zone)
        kwargs.setdefault("subnet_id", stack.default_subnet_id)
        kwargs.setdefault("root_device_type", stack.default_root_device_type)
        if layer.custom_instance_profile_arn:
            kwargs.setdefault("instance_profile_arn", layer.custom_instance_profile_arn)
        kwargs.setdefault("instance_profile_arn", stack.default_instance_profile_arn)
        kwargs.setdefault("security_group_ids", layer.custom_security_group_ids)
        kwargs.setdefault("associate_public_ip", layer.auto_assign_public_ips)
        kwargs.setdefault("ebs_optimized", layer.use_ebs_optimized_instances)
        kwargs.update({"ec2_backend": self.ec2_backend})
        opsworks_instance = OpsworkInstance(**kwargs)
        self.instances[opsworks_instance.id] = opsworks_instance
        return opsworks_instance

    def describe_stacks(self, stack_ids):
        if stack_ids is None:
            return [stack.to_dict() for stack in self.stacks.values()]

        unknown_stacks = set(stack_ids) - set(self.stacks.keys())
        if unknown_stacks:
            raise ResourceNotFoundException(", ".join(unknown_stacks))
        return [self.stacks[id].to_dict() for id in stack_ids]

    def describe_layers(self, stack_id, layer_ids):
        if stack_id is not None and layer_ids is not None:
            raise ValidationException(
                "Please provide one or more layer IDs or a stack ID"
            )
        if stack_id is not None:
            if stack_id not in self.stacks:
                raise ResourceNotFoundException(
                    "Unable to find stack with ID {0}".format(stack_id)
                )
            return [layer.to_dict() for layer in self.stacks[stack_id].layers]

        unknown_layers = set(layer_ids) - set(self.layers.keys())
        if unknown_layers:
            raise ResourceNotFoundException(", ".join(unknown_layers))
        return [self.layers[id].to_dict() for id in layer_ids]

    def describe_apps(self, stack_id, app_ids):
        if stack_id is not None and app_ids is not None:
            raise ValidationException(
                "Please provide one or more app IDs or a stack ID"
            )
        if stack_id is not None:
            if stack_id not in self.stacks:
                raise ResourceNotFoundException(
                    "Unable to find stack with ID {0}".format(stack_id)
                )
            return [app.to_dict() for app in self.stacks[stack_id].apps]

        unknown_apps = set(app_ids) - set(self.apps.keys())
        if unknown_apps:
            raise ResourceNotFoundException(", ".join(unknown_apps))
        return [self.apps[id].to_dict() for id in app_ids]

    def describe_instances(self, instance_ids, layer_id, stack_id):
        if len(list(filter(None, (instance_ids, layer_id, stack_id)))) != 1:
            raise ValidationException(
                "Please provide either one or more "
                "instance IDs or one stack ID or one "
                "layer ID"
            )
        if instance_ids:
            unknown_instances = set(instance_ids) - set(self.instances.keys())
            if unknown_instances:
                raise ResourceNotFoundException(", ".join(unknown_instances))
            return [self.instances[id].to_dict() for id in instance_ids]

        if layer_id:
            if layer_id not in self.layers:
                raise ResourceNotFoundException(
                    "Unable to find layer with ID {0}".format(layer_id)
                )
            instances = [
                i.to_dict() for i in self.instances.values() if layer_id in i.layer_ids
            ]
            return instances

        if stack_id:
            if stack_id not in self.stacks:
                raise ResourceNotFoundException(
                    "Unable to find stack with ID {0}".format(stack_id)
                )
            instances = [
                i.to_dict() for i in self.instances.values() if stack_id == i.stack_id
            ]
            return instances

    def start_instance(self, instance_id):
        if instance_id not in self.instances:
            raise ResourceNotFoundException(
                "Unable to find instance with ID {0}".format(instance_id)
            )
        self.instances[instance_id].start()


opsworks_backends = BackendDict(OpsWorksBackend, "ec2")



import boto3
import sure  # noqa # pylint: disable=unused-import

from moto import mock_autoscaling, mock_elb, mock_ec2

from .utils import setup_networking
from tests import EXAMPLE_AMI_ID


@mock_autoscaling
@mock_ec2
@mock_elb
def test_enable_metrics_collection():
    mocked_networking = setup_networking()
    elb_client = boto3.client("elb", region_name="us-east-1")
    elb_client.create_load_balancer(
        LoadBalancerName="test_lb",
        Listeners=[{"Protocol": "http", "LoadBalancerPort": 80, "InstancePort": 8080}],
        AvailabilityZones=[],
    )

    as_client = boto3.client("autoscaling", region_name="us-east-1")
    as_client.create_launch_configuration(
        LaunchConfigurationName="tester_config",
        ImageId=EXAMPLE_AMI_ID,
        InstanceType="t2.medium",
    )

    as_client.create_auto_scaling_group(
        AutoScalingGroupName="tester_group",
        LaunchConfigurationName="tester_config",
        MinSize=2,
        MaxSize=2,
        VPCZoneIdentifier=mocked_networking["subnet1"],
    )

    as_client.enable_metrics_collection(
        AutoScalingGroupName="tester_group",
        Metrics=["GroupMinSize"],
        Granularity="1Minute",
    )

    resp = as_client.describe_auto_scaling_groups(
        AutoScalingGroupNames=["tester_group"]
    )["AutoScalingGroups"][0]
    resp.should.have.key("EnabledMetrics").length_of(1)
    resp["EnabledMetrics"][0].should.equal(
        {"Metric": "GroupMinSize", "Granularity": "1Minute"}
    )
